"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import sys
import logging
import functools
from . import mpv2 as mpv
import config as cfg
import gui_utilities


from PySide6.QtWidgets import (
    QLabel,
    QDockWidget,
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QSlider,
    QSizePolicy,
    QStackedWidget,
    QToolButton,
)
from PySide6.QtCore import Signal, QEvent, Qt
from PySide6.QtGui import QIcon, QAction


class Clickable_label(QLabel):
    """
    QLabel class for visualiziong frames for geometric measurments
    Label emits a signal when clicked
    """

    mouse_pressed_signal = Signal(int, QEvent)

    def __init__(self, id_, parent=None):
        QLabel.__init__(self, parent)
        self.id_ = id_

    def mousePressEvent(self, event):
        """
        label clicked
        """
        logging.debug(f"mousepress event: label {self.id_} clicked {event.pos()}")

        """
        super().mousePressEvent(event)
        x, y = event.pos().x(), event.pos().y()
        draw_on_pixmap(self, x, y)  # Example usage
        """

        self.mouse_pressed_signal.emit(self.id_, event)


def mpv_logger(player_id, loglevel, component, message):
    """
    redirect MPV log messages to general logging system
    """

    logging.debug(f"MPV player #{player_id}: [{loglevel}] {component}: {message}")


class DW_player(QDockWidget):
    """
    Define the player class
    """

    key_pressed_signal = Signal(QEvent)
    volume_slider_moved_signal = Signal(int, int)
    mute_action_triggered_signal = Signal(int)
    view_signal = Signal(int, str, int)
    resize_signal = Signal(int)

    def __init__(self, id_, parent=None):
        super().__init__(parent)
        self.id_ = id_
        self.setWindowTitle(f"Player #{id_ + 1}")
        self.setObjectName(f"player{id_ + 1}")

        self.stack1 = QWidget()
        self.hlayout = QHBoxLayout()

        self.videoframe = QWidget(self)

        if sys.platform.startswith(cfg.MACOS_CODE):
            self.player = None
        else:
            self.player = mpv.MPV(
                wid=str(int(self.videoframe.winId())),
                vo="x11" if sys.platform.startswith("linux") else "",
                log_handler=functools.partial(mpv_logger, self.id_),
                loglevel="debug",
            )

        self.player.screenshot_format = "png"
        self.hlayout.addWidget(self.videoframe)

        # layout volume slider and mute button
        self.vlayout = QVBoxLayout()

        # volume slider
        self.volume_slider = QSlider(Qt.Vertical, self)
        self.volume_slider.setFocusPolicy(Qt.NoFocus)
        self.volume_slider.setMaximum(100)
        self.volume_slider.setValue(50)
        self.volume_slider.sliderMoved.connect(self.volume_slider_moved)

        self.vlayout.addWidget(self.volume_slider)

        # mute button
        self.mute_button = QToolButton()
        self.mute_button.setFocusPolicy(Qt.NoFocus)
        self.mute_button.setAutoRaise(True)
        self.mute_action = QAction()

        theme_mode = gui_utilities.theme_mode()

        self.mute_action.setIcon(QIcon(f":/volume_xmark_{theme_mode}"))
        self.mute_action.triggered.connect(self.mute_action_triggered)
        self.mute_button.setDefaultAction(self.mute_action)

        self.vlayout.addWidget(self.mute_button)

        self.hlayout.addLayout(self.vlayout)

        self.stack1.setLayout(self.hlayout)

        self.stack2 = QWidget()
        self.hlayout2 = QHBoxLayout()
        self.frame_viewer = Clickable_label(id_)

        self.frame_viewer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.frame_viewer.setAlignment(Qt.AlignCenter)
        self.frame_viewer.setStyleSheet("QLabel {background-color: black;}")

        self.hlayout2.addWidget(self.frame_viewer)
        self.stack2.setLayout(self.hlayout2)

        self.stack = QStackedWidget(self)
        self.stack.addWidget(self.stack1)
        self.stack.addWidget(self.stack2)

        self.setWidget(self.stack)

        self.stack.setCurrentIndex(0)

    def volume_slider_moved(self):
        """
        emit signal when volume slider moved
        """
        self.volume_slider_moved_signal.emit(self.id_, self.volume_slider.value())

    def mute_action_triggered(self):
        """
        emit signal when mute action is triggered
        """
        theme_mode = gui_utilities.theme_mode()
        if self.player.mute:
            self.mute_action.setIcon(QIcon(f":/volume_xmark_{theme_mode}"))
        else:
            self.mute_action.setIcon(QIcon(f":/volume_off_{theme_mode}"))
        self.mute_action_triggered_signal.emit(self.id_)

    def keyPressEvent(self, event):
        """
        emit signal when key pressed on dock widget
        """
        self.key_pressed_signal.emit(event)

    def resizeEvent(self, _):
        """
        emits signal when dockwidget resized
        """

        self.resize_signal.emit(self.id_)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import pathlib as pl
import logging
from PySide6.QtCore import QSettings
from PySide6.QtWidgets import QWidget, QApplication
from PySide6.QtGui import QIcon


def theme_mode() -> str:
    """
    return the theme mode (dark or light) of the OS
    """
    palette = QApplication.instance().palette()
    color = palette.window().color()
    return "dark" if color.value() < 128 else "light"  # Dark mode if the color value is less than 128


def save_geometry(widget: QWidget, widget_name: str):
    """
    save window geometry in ini file
    """

    ini_file_path = pl.Path.home() / pl.Path(".boris")
    if ini_file_path.is_file():
        try:
            settings = QSettings(str(ini_file_path), QSettings.IniFormat)
            settings.setValue(f"{widget_name} geometry", widget.saveGeometry())
        except Exception:
            logging.warning(f"error during saving {widget_name} geometry")


def restore_geometry(widget: QWidget, widget_name: str, default_width_height):
    """
    restore window geometry in ini file
    """

    def default_resize(widget, default_width_height):
        if default_width_height != (0, 0):
            try:
                widget.resize(default_width_height[0], default_width_height[1])
            except Exception:
                logging.warning("Error during restoring default")

    logging.debug(f"restore geometry function for {widget_name}")
    try:
        ini_file_path = pl.Path.home() / pl.Path(".boris")
        if ini_file_path.is_file():
            settings = QSettings(str(ini_file_path), QSettings.IniFormat)
            print(settings.value(f"{widget_name} geometry"))
            widget.restoreGeometry(settings.value(f"{widget_name} geometry"))
            logging.debug(f"geometry restored for {widget_name}  {settings.value(f'{widget_name} geometry')}")
        else:
            default_resize(widget, default_width_height)
    except Exception:
        logging.warning(f"error during restoring {widget_name} geometry")
        default_resize(widget, default_width_height)


def set_icons(self, theme_mode: str) -> None:
    """
    set icons of actions
    """

    # menu
    self.action_obs_list.setIcon(QIcon(f":/observations_list_{theme_mode}"))

    self.actionTime_budget.setIcon(QIcon(f":/time_budget_{theme_mode}"))
    self.actionPlot_events2.setIcon(QIcon(f":/plot_events_{theme_mode}"))
    self.action_advanced_event_filtering.setIcon(QIcon(f":/filter_{theme_mode}"))

    self.actionPreferences.setIcon(QIcon(f":/preferences_{theme_mode}"))

    self.actionPlay.setIcon(QIcon(f":/play_{theme_mode}"))
    self.actionReset.setIcon(QIcon(f":/reset_{theme_mode}"))
    self.actionJumpBackward.setIcon(QIcon(f":/jump_backward_{theme_mode}"))
    self.actionJumpForward.setIcon(QIcon(f":/jump_forward_{theme_mode}"))

    self.actionFaster.setIcon(QIcon(f":/faster_{theme_mode}"))
    self.actionSlower.setIcon(QIcon(f":/slower_{theme_mode}"))
    self.actionNormalSpeed.setIcon(QIcon(f":/normal_speed_{theme_mode}"))

    self.actionPrevious.setIcon(QIcon(f":/previous_{theme_mode}"))
    self.actionNext.setIcon(QIcon(f":/next_{theme_mode}"))

    self.actionSnapshot.setIcon(QIcon(f":/snapshot_{theme_mode}"))

    self.actionFrame_backward.setIcon(QIcon(f":/frame_backward_{theme_mode}"))
    self.actionFrame_forward.setIcon(QIcon(f":/frame_forward_{theme_mode}"))
    self.actionCloseObs.setIcon(QIcon(f":/close_observation_{theme_mode}"))
    self.actionCurrent_Time_Budget.setIcon(QIcon(f":/time_budget_{theme_mode}"))
    self.actionPlot_current_observation.setIcon(QIcon(f":/plot_events_{theme_mode}"))

    self.actionPlot_events_in_real_time.setIcon(QIcon(f":/plot_real_time_{theme_mode}"))

    self.actionBehavior_bar_plot.setIcon(QIcon(f":/plot_time_budget_{theme_mode}"))
    self.actionPlot_current_time_budget.setIcon(QIcon(f":/plot_time_budget_{theme_mode}"))
    self.action_geometric_measurements.setIcon(QIcon(f":/measurement_{theme_mode}"))
    self.actionFind_in_current_obs.setIcon(QIcon(f":/find_{theme_mode}"))
    self.actionExplore_project.setIcon(QIcon(f":/explore_{theme_mode}"))


def resize_center(app, window, width: int, height: int) -> None:
    """
    resize and center window
    """
    window.resize(width, height)
    screen_geometry = app.primaryScreen().geometry()
    if window.height() > screen_geometry.height():
        window.resize(window.width(), int(screen_geometry.height() * 0.8))
    if window.width() > screen_geometry.width():
        window.resize(screen_geometry.width(), window.height())
    # center
    center_x = (screen_geometry.width() - window.width()) // 2
    center_y = (screen_geometry.height() - window.height()) // 2

    window.move(center_x, center_y)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import binascii

from PySide6.QtCore import Signal, QPoint, Qt
from PySide6.QtGui import QPen, QPixmap, QBrush, QMouseEvent, QPolygonF, QColor
from PySide6.QtWidgets import (
    QDialog,
    QGraphicsView,
    QGraphicsScene,
    QLineEdit,
    QVBoxLayout,
    QHBoxLayout,
    QSpacerItem,
    QSizePolicy,
    QPushButton,
    QGraphicsPixmapItem,
    QGraphicsPolygonItem,
)


class ModifiersCodingMapWindow(QDialog):
    class View(QGraphicsView):
        mousePress = Signal(QMouseEvent)

        def mousePressEvent(self, event):
            self.mousePress.emit(event)

        start = 0
        elList = []
        points = []

        def __init__(self, parent):
            QGraphicsView.__init__(self, parent)
            self.setScene(QGraphicsScene(self))
            self.scene().update()

    def __init__(self, modifiers_coding_map):
        super(ModifiersCodingMapWindow, self).__init__()

        self.areasList = {}
        self.polygonsList2 = {}
        self.codeSeparator = ","

        self.codingMap = modifiers_coding_map
        self.setWindowTitle(f"Modifiers coding map: {self.codingMap['name']}")
        Vlayout = QVBoxLayout(self)

        self.view = self.View(self)
        self.view.mousePress.connect(self.viewMousePressEvent)

        self.leareaCode = QLineEdit(self)
        self.leareaCode.setVisible(True)

        Vlayout.addWidget(self.view)
        Vlayout.addWidget(self.leareaCode)

        hBoxLayout = QHBoxLayout(self)

        spacerItem = QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        hBoxLayout.addItem(spacerItem)

        self.pbCancel = QPushButton("Cancel", clicked=self.reject)
        hBoxLayout.addWidget(self.pbCancel)

        self.btDone = QPushButton("Done", self, clicked=self.accept)
        self.btDone.setVisible(True)

        hBoxLayout.addWidget(self.btDone)

        Vlayout.addLayout(hBoxLayout)

        self.setLayout(Vlayout)

        self.loadMap()

    def viewMousePressEvent(self, event):
        """
        insert clicked areas codes
        """

        test = self.view.mapToScene(event.pos()).toPoint()

        for code in self.polygonsList2:
            if self.polygonsList2[code].contains(test):
                codes = self.leareaCode.text().split(self.codeSeparator)
                if "" in codes:
                    codes.remove("")

                # check if code already in codes list
                if code in codes:
                    codes.remove(code)
                else:
                    codes.append(code)

                self.leareaCode.setText(self.codeSeparator.join(sorted(codes)))

    def getCodes(self):
        return self.leareaCode.text()

    def loadMap(self):
        """
        load bitmap from data
        show it in view scene
        """

        self.areasList = self.codingMap

        bitmapContent = binascii.a2b_base64(self.areasList["bitmap"])

        pixmap = QPixmap()
        pixmap.loadFromData(bitmapContent)

        self.view.setSceneRect(0, 0, pixmap.size().width(), pixmap.size().height())
        pixItem = QGraphicsPixmapItem(pixmap)
        pixItem.setPos(0, 0)
        self.view.scene().addItem(pixItem)

        for area in self.areasList["areas"]:
            points = self.areasList["areas"][area]["geometry"]

            newPolygon = QPolygonF()
            for p in points:
                newPolygon.append(QPoint(p[0], p[1]))

            clr = QColor()
            clr.setRgba(self.areasList["areas"][area]["color"])

            # draw polygon
            polygon = QGraphicsPolygonItem(newPolygon)

            polygon.setPen(QPen(clr, 0, Qt.NoPen, Qt.RoundCap, Qt.RoundJoin))

            polygon.setBrush(QBrush(clr, Qt.SolidPattern))

            self.view.scene().addItem(polygon)
            self.polygonsList2[area] = polygon



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

from PySide6.QtCore import Qt, QTimer
from PySide6.QtGui import QAction

from . import config as cfg
from . import (
    about,
    behav_coding_map_creator,
    behavior_binary_table,
    behaviors_coding_map,
    coding_pad,
    event_operations,
    events_snapshots,
    export_events,
    geometric_measurement,
    image_overlay,
    import_observations,
    irr,
    latency,
    cooccurence,
    media_file,
    observation_operations,
    preferences,
    project_import_export,
    synthetic_time_budget,
    time_budget_widget,
    transitions,
    video_equalizer,
    video_operations,
    project_functions,
    external_processes,
)

from . import state_events as state_events


def connections(self):
    """
    create connections between widgets and functions
    """

    # menu file
    self.actionNew_project.triggered.connect(lambda: self.edit_project(cfg.NEW))
    self.actionOpen_project.triggered.connect(self.open_project_activated)
    self.actionNoldus_Observer_template.triggered.connect(self.import_project_from_observer_template)
    self.actionEdit_project.triggered.connect(self.edit_project_activated)
    self.actionCheck_project.triggered.connect(self.check_project_integrity)
    self.actionSave_project.triggered.connect(self.save_project_activated)
    self.actionSave_project_as.triggered.connect(self.save_project_as_activated)
    self.actionExport_project.triggered.connect(lambda: project_import_export.export_project_as_pickle_object(self.pj))
    self.actionClose_project.triggered.connect(self.close_project)

    self.action_media_file_and_images_directories_relative_path.triggered.connect(self.set_media_files_path_relative_to_project_dir)
    self.action_data_files_relative_path.triggered.connect(self.set_data_files_path_relative_to_project_dir)

    self.action_remove_media_files_and_images_directories_path.triggered.connect(self.remove_media_files_path)
    self.action_remove_data_files_path.triggered.connect(self.remove_data_files_path)

    self.menuCreate_subtitles_2.triggered.connect(self.create_subtitles)

    self.actionPreferences.triggered.connect(lambda: preferences.preferences(self))

    self.actionQuit.triggered.connect(self.actionQuit_activated)

    # menu observations
    self.actionNew_observation.triggered.connect(lambda: observation_operations.new_observation(self, mode=cfg.NEW, obsId=""))

    self.actionOpen_observation.triggered.connect(lambda: observation_operations.open_observation(self, mode=cfg.OBS_START))
    self.actionView_observation.triggered.connect(lambda: observation_operations.open_observation(self, mode=cfg.VIEW))
    self.actionEdit_observation_2.triggered.connect(lambda: observation_operations.edit_observation(self))
    self.actionObservationsList.triggered.connect(lambda: observation_operations.observations_list(self))

    self.actionClose_observation.triggered.connect(lambda: observation_operations.close_observation(self))

    self.action_create_observations.triggered.connect(lambda: observation_operations.create_observations(self))
    self.actionRemove_observations.triggered.connect(lambda: observation_operations.remove_observations(self))

    self.actionAdd_event.triggered.connect(lambda: event_operations.add_event(self))
    self.actionEdit_event.triggered.connect(lambda: event_operations.edit_event(self))
    self.actionUndo.setEnabled(False)
    self.actionUndo.triggered.connect(lambda: event_operations.undo_event_operation(self))
    self.actionFilter_events.triggered.connect(lambda: event_operations.filter_events(self))
    self.actionShow_all_events.triggered.connect(lambda: event_operations.show_all_events(self))

    # twevent header
    # self.actionConfigure_twEvents_columns.triggered.connect(self.configure_twevents_columns)

    # tv_events header
    self.actionConfigure_tvevents_columns.triggered.connect(self.configure_tvevents_columns)

    self.actionExport_observations_list.triggered.connect(lambda: observation_operations.export_observations_list_clicked(self))

    self.actionCheckStateEvents.triggered.connect(lambda: state_events.check_state_events(self, mode="all"))
    self.actionCheckStateEventsSingleObs.triggered.connect(lambda: state_events.check_state_events(self, mode="current"))
    self.actionClose_unpaired_events.triggered.connect(lambda: state_events.fix_unpaired_events(self))
    self.actionAdd_frame_indexes.triggered.connect(lambda: event_operations.add_frame_indexes(self))

    self.actionRunEventOutside.triggered.connect(self.run_event_outside)

    self.actionSelect_observations.triggered.connect(lambda: event_operations.select_events_between_activated(self))

    self.actionEdit_selected_events.triggered.connect(lambda: event_operations.edit_selected_events(self))
    self.actionEdit_event_time.triggered.connect(lambda: event_operations.edit_time_selected_events(self))

    self.actionCopy_events.triggered.connect(lambda: event_operations.copy_selected_events(self))
    self.actionPaste_events.triggered.connect(lambda: event_operations.paste_clipboard_to_events(self))

    self.actionExplore_project.triggered.connect(lambda: project_functions.explore_project(self))
    self.actionFind_events.triggered.connect(lambda: event_operations.find_events(self))
    self.actionFind_replace_events.triggered.connect(lambda: event_operations.find_replace_events(self))
    self.actionDelete_all_events.triggered.connect(lambda: event_operations.delete_all_events(self))
    self.actionDelete_selected_events.triggered.connect(lambda: event_operations.delete_selected_events(self))

    self.actionMedia_file_information.triggered.connect(lambda: media_file.get_info(self))

    self.actionLoad_observations_file.triggered.connect(lambda: import_observations.import_observations(self))

    self.actionExportEvents_2.triggered.connect(lambda: export_events.export_tabular_events(self, mode="tabular"))

    # behavioral sequences
    # self.actionExportEventString.triggered.connect(lambda: self.export_events_as_behavioral_sequences(timed=False))
    self.actionseparated_subjects.triggered.connect(
        lambda: export_events.export_events_as_behavioral_sequences(self, separated_subjects=True, timed=False)
    )
    self.actiongrouped_subjects.triggered.connect(
        lambda: export_events.export_events_as_behavioral_sequences(self, separated_subjects=False, timed=False)
    )

    self.actionExport_aggregated_events.triggered.connect(lambda: export_events.export_aggregated_events(self))
    self.actionExport_events_as_Praat_TextGrid.triggered.connect(lambda: export_events.export_events_as_textgrid(self))
    self.actionJWatcher.triggered.connect(lambda: export_events.export_tabular_events(self, "jwatcher"))

    self.actionExtract_events_from_media_files.triggered.connect(lambda: events_snapshots.extract_events(self))
    self.actionExtract_frames_from_media_files.triggered.connect(lambda: events_snapshots.events_snapshots(self))

    self.actionCohen_s_kappa.triggered.connect(lambda: irr.irr_cohen_kappa(self))
    self.actionNeedleman_Wunsch.triggered.connect(lambda: irr.needleman_wunch(self))

    self.actionAll_transitions.triggered.connect(lambda: transitions.transitions_matrix(self, "frequency"))
    self.actionNumber_of_transitions.triggered.connect(lambda: transitions.transitions_matrix(self, "number"))

    self.actionFrequencies_of_transitions_after_behaviors.triggered.connect(
        lambda: transitions.transitions_matrix(self, "frequencies_after_behaviors")
    )

    # menu playback
    self.actionJumpTo.triggered.connect(self.jump_to)
    self.action_deinterlace.triggered.connect(lambda: video_operations.deinterlace(self))
    self.actionZoom_level.triggered.connect(lambda: video_operations.zoom_level(self))
    self.actionRotate_current_video.triggered.connect(lambda: video_operations.rotate_displayed_video(self))
    self.actionDisplay_subtitles.triggered.connect(lambda: video_operations.display_subtitles(self))
    self.actionVideo_equalizer.triggered.connect(lambda: video_equalizer.video_equalizer_show(self))

    # menu Tools
    self.action_block_dockwidgets.triggered.connect(self.block_dockwidgets)

    self.action_create_modifiers_coding_map.triggered.connect(self.modifiers_coding_map_creator)
    self.action_create_behaviors_coding_map.triggered.connect(lambda: behav_coding_map_creator.behaviors_coding_map_creator(self))

    self.actionShow_spectrogram.triggered.connect(lambda: self.show_plot_widget("spectrogram", warning=True))
    self.actionShow_the_sound_waveform.triggered.connect(lambda: self.show_plot_widget("waveform", warning=True))
    self.actionPlot_events_in_real_time.triggered.connect(lambda: self.show_plot_widget("plot_events", warning=False))

    self.actionShow_data_files.triggered.connect(self.show_data_files)
    self.action_geometric_measurements.triggered.connect(lambda: geometric_measurement.show_widget(self))

    self.actionBehaviors_coding_map.triggered.connect(lambda: behaviors_coding_map.show_behaviors_coding_map(self))

    self.actionCoding_pad.triggered.connect(lambda: coding_pad.show_coding_pad(self))
    self.actionSubjects_pad.triggered.connect(self.show_subjects_pad)

    # image overlay on video
    self.actionAdd_image_overlay_on_video.triggered.connect(lambda: image_overlay.add_image_overlay(self))
    self.actionRemove_image_overlay.triggered.connect(lambda: image_overlay.remove_image_overlay(self))

    self.actionRecode_resize_video.triggered.connect(lambda: external_processes.ffmpeg_process(self, "reencode_resize"))
    self.actionRotate_video.triggered.connect(lambda: external_processes.ffmpeg_process(self, "rotate"))
    self.actionMerge_media_files.triggered.connect(lambda: external_processes.ffmpeg_process(self, "merge"))
    self.actionMedia_file_information_2.triggered.connect(lambda: media_file.get_info(self))

    self.actionCreate_transitions_flow_diagram.triggered.connect(transitions.transitions_dot_script)
    self.actionCreate_transitions_flow_diagram_2.triggered.connect(transitions.transitions_flow_diagram)

    # menu Analysis

    self.actionTime_budget.triggered.connect(lambda: time_budget_widget.time_budget(self, mode="by_behavior"))
    self.actionTime_budget_by_behaviors_category.triggered.connect(lambda: time_budget_widget.time_budget(self, mode="by_category"))

    self.actionTime_budget_report.triggered.connect(lambda: synthetic_time_budget.synthetic_time_budget(self))
    self.actionSynthetic_binned_time_budget.triggered.connect(lambda: synthetic_time_budget.synthetic_binned_time_budget(self))

    self.actionBehavior_bar_plot.triggered.connect(lambda: self.behaviors_bar_plot(mode="list"))
    self.actionBehavior_bar_plot.setVisible(True)

    self.actionPlot_events1.setVisible(False)
    self.actionPlot_events2.triggered.connect(lambda: self.plot_events_triggered(mode="list"))

    self.action_behavior_binary_table.triggered.connect(lambda: behavior_binary_table.behavior_binary_table(self))

    self.action_advanced_event_filtering.triggered.connect(self.advanced_event_filtering)

    self.action_latency.triggered.connect(lambda: latency.get_latency(self))

    self.action_cooccurence.triggered.connect(lambda: cooccurence.get_cooccurence(self))

    # menu Help
    self.actionUser_guide.triggered.connect(self.actionUser_guide_triggered)
    self.actionAbout.triggered.connect(lambda: about.actionAbout_activated(self))
    self.actionCheckUpdate.triggered.connect(self.actionCheckUpdate_activated)

    # toolbar
    self.action_obs_list.triggered.connect(lambda: observation_operations.observations_list(self))
    self.actionPlay.triggered.connect(self.play_activated)
    self.actionReset.triggered.connect(self.reset_activated)
    self.actionJumpBackward.triggered.connect(self.jumpBackward_activated)
    self.actionJumpForward.triggered.connect(self.jumpForward_activated)

    self.actionFaster.triggered.connect(lambda: video_operations.video_faster_activated(self))
    self.actionSlower.triggered.connect(lambda: video_operations.video_slower_activated(self))
    self.actionNormalSpeed.triggered.connect(lambda: video_operations.video_normalspeed_activated(self))

    self.actionPrevious.triggered.connect(self.previous_media_file)
    self.actionNext.triggered.connect(self.next_media_file)

    self.actionSnapshot.triggered.connect(lambda: video_operations.snapshot(self))

    self.actionFrame_backward.triggered.connect(self.previous_frame)
    self.actionFrame_forward.triggered.connect(self.next_frame)

    self.actionCloseObs.triggered.connect(lambda: observation_operations.close_observation(self))
    self.actionCurrent_Time_Budget.triggered.connect(lambda: time_budget_widget.time_budget(self, mode="by_behavior", mode2="current"))
    self.actionPlot_current_observation.triggered.connect(lambda: self.plot_events_triggered(mode="current"))

    self.actionPlot_current_time_budget.triggered.connect(lambda: self.behaviors_bar_plot(mode="current"))

    self.actionFind_in_current_obs.triggered.connect(lambda: event_operations.find_events(self))

    # table Widget double click
    # self.twEvents.itemDoubleClicked.connect(self.twEvents_doubleClicked)
    self.twEthogram.itemDoubleClicked.connect(self.twEthogram_doubleClicked)
    self.twSubjects.itemDoubleClicked.connect(self.twSubjects_doubleClicked)

    # events tableview
    self.tv_events.doubleClicked.connect(self.tv_events_doubleClicked)

    # Actions for twEthogram context menu
    self.twEthogram.setContextMenuPolicy(Qt.ActionsContextMenu)
    self.twEthogram.horizontalHeader().sortIndicatorChanged.connect(self.twEthogram_sorted)

    self.actionViewBehavior.triggered.connect(self.view_behavior)
    self.twEthogram.addAction(self.actionViewBehavior)

    self.actionFilterBehaviors.triggered.connect(
        lambda: self.filter_behaviors(table=cfg.ETHOGRAM, behavior_type=cfg.STATE_EVENT_TYPES + cfg.POINT_EVENT_TYPES)
    )
    self.twEthogram.addAction(self.actionFilterBehaviors)

    self.actionShowAllBehaviors.triggered.connect(self.show_all_behaviors)
    self.twEthogram.addAction(self.actionShowAllBehaviors)

    # Actions for twSubjects context menu
    self.twSubjects.setContextMenuPolicy(Qt.ActionsContextMenu)
    self.twSubjects.horizontalHeader().sortIndicatorChanged.connect(self.sort_twSubjects)
    self.actionFilterSubjects.triggered.connect(self.filter_subjects)
    self.twSubjects.addAction(self.actionFilterSubjects)

    self.actionShowAllSubjects.triggered.connect(self.show_all_subjects)
    self.twSubjects.addAction(self.actionShowAllSubjects)

    # actions for twEvents horizontal header menu
    # tw_headers = self.twEvents.horizontalHeader()
    # tw_headers.setContextMenuPolicy(Qt.ActionsContextMenu)
    # tw_headers.addAction(self.actionConfigure_twEvents_columns)

    tv_headers = self.tv_events.horizontalHeader()
    tv_headers.setContextMenuPolicy(Qt.ActionsContextMenu)
    tv_headers.addAction(self.actionConfigure_tvevents_columns)

    # Actions for twEvents menu
    # self.twEvents.setContextMenuPolicy(Qt.ActionsContextMenu)

    # self.twEvents.addAction(self.actionAdd_event)
    # self.twEvents.addAction(self.actionEdit_selected_events)
    # self.twEvents.addAction(self.actionEdit_event_time)

    # self.twEvents.addAction(self.actionCopy_events)
    # self.twEvents.addAction(self.actionPaste_events)

    # separator2 = QAction(self)
    # separator2.setSeparator(True)
    # self.twEvents.addAction(separator2)

    # self.twEvents.addAction(self.actionFind_events)
    # self.twEvents.addAction(self.actionFind_replace_events)

    # separator2 = QAction(self)
    # separator2.setSeparator(True)
    # self.twEvents.addAction(separator2)

    # self.twEvents.addAction(self.actionFilter_events)
    # self.twEvents.addAction(self.actionShow_all_events)

    # separator2 = QAction(self)
    # separator2.setSeparator(True)
    # self.twEvents.addAction(separator2)

    # self.twEvents.addAction(self.actionCheckStateEventsSingleObs)
    # self.twEvents.addAction(self.actionClose_unpaired_events)

    # self.twEvents.addAction(self.actionRunEventOutside)

    # separator2 = QAction(self)
    # separator2.setSeparator(True)
    # self.twEvents.addAction(separator2)

    # self.twEvents.addAction(self.actionDelete_selected_events)

    # Actions for tv_events menu
    self.tv_events.setContextMenuPolicy(Qt.ActionsContextMenu)

    self.tv_events.addAction(self.actionAdd_event)
    self.tv_events.addAction(self.actionEdit_selected_events)
    self.tv_events.addAction(self.actionEdit_event_time)

    self.tv_events.addAction(self.actionCopy_events)
    self.tv_events.addAction(self.actionPaste_events)

    separator2 = QAction(self)
    separator2.setSeparator(True)
    self.tv_events.addAction(separator2)

    self.tv_events.addAction(self.actionFind_events)
    self.tv_events.addAction(self.actionFind_replace_events)

    separator2 = QAction(self)
    separator2.setSeparator(True)
    self.tv_events.addAction(separator2)

    self.tv_events.addAction(self.actionFilter_events)
    self.tv_events.addAction(self.actionShow_all_events)

    separator2 = QAction(self)
    separator2.setSeparator(True)
    self.tv_events.addAction(separator2)

    self.tv_events.addAction(self.actionCheckStateEventsSingleObs)
    self.tv_events.addAction(self.actionClose_unpaired_events)

    self.tv_events.addAction(self.actionAdd_frame_indexes)

    self.tv_events.addAction(self.actionRunEventOutside)

    separator2 = QAction(self)
    separator2.setSeparator(True)
    self.tv_events.addAction(separator2)

    self.tv_events.addAction(self.actionDelete_selected_events)

    # Actions for twSubjects context menu
    self.actionDeselectCurrentSubject.triggered.connect(lambda: self.update_subject(""))

    self.twSubjects.setContextMenuPolicy(Qt.ActionsContextMenu)
    self.twSubjects.addAction(self.actionDeselectCurrentSubject)

    # subjects

    # timer for plot visualization
    self.plot_timer = QTimer(self)
    # TODO check value of interval
    self.plot_timer.setInterval(cfg.SPECTRO_TIMER)
    self.plot_timer.timeout.connect(self.plot_timer_out)

    # timer for timing the live observation
    self.liveTimer = QTimer(self)
    self.liveTimer.timeout.connect(self.liveTimer_out)

    # timer for automatic backup
    self.automaticBackupTimer = QTimer(self)
    self.automaticBackupTimer.timeout.connect(self.automatic_backup)
    if self.automaticBackup:
        self.automaticBackupTimer.start(self.automaticBackup * 60000)

    self.pb_live_obs.clicked.connect(self.start_live_observation)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import urllib
import json
import pathlib as pl
import pandas as pd
import tablib
import pickle

from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QApplication, QFileDialog, QListWidgetItem, QMessageBox, QTableWidgetItem


from . import config as cfg
from . import dialog, param_panel, project_functions, export_observation
from . import utilities as util


def export_project_as_pickle_object(pj: dict) -> None:
    """
    export the project dictionary as a pickle file
    """
    file_name, _ = QFileDialog().getSaveFileName(None, "Export project as pickle file", "", "All files (*)")
    if not file_name:
        return
    try:
        with open(file_name, "wb") as f_out:
            pickle.dump(pj, f_out)
    except Exception:
        QMessageBox.critical(
            None,
            cfg.programName,
            "Error during file saving.",
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )


def export_ethogram(self) -> None:
    """
    export ethogram in various format
    """
    extended_file_formats: list = [
        "BORIS project file (*.boris)",
        "Tab Separated Values (*.tsv)",
        "Comma Separated Values (*.csv)",
        "Open Document Spreadsheet ODS (*.ods)",
        "Microsoft Excel Spreadsheet XLSX (*.xlsx)",
        "Legacy Microsoft Excel Spreadsheet XLS (*.xls)",
        "HTML (*.html)",
    ]
    file_formats: list = ["boris", cfg.TSV_EXT, cfg.CSV_EXT, cfg.ODS_EXT, cfg.XLSX_EXT, cfg.XLS_EXT, cfg.HTML_EXT]

    filediag_func = QFileDialog().getSaveFileName

    file_name, filter_ = filediag_func(self, "Export ethogram", "", ";;".join(extended_file_formats))
    if not file_name:
        return

    output_format: str = file_formats[extended_file_formats.index(filter_)]
    if pl.Path(file_name).suffix != "." + output_format:
        file_name = str(pl.Path(file_name)) + "." + output_format

    if output_format == "boris":
        r = self.check_ethogram()
        if cfg.CANCEL in r:
            return
        pj = dict(cfg.EMPTY_PROJECT)
        pj[cfg.ETHOGRAM] = dict(r)

        # behavioral categories
        pj[cfg.BEHAVIORAL_CATEGORIES] = list(self.pj[cfg.BEHAVIORAL_CATEGORIES])
        pj[cfg.BEHAVIORAL_CATEGORIES_CONF] = dict(self.pj.get(cfg.BEHAVIORAL_CATEGORIES_CONF, {}))

        # project file indentation
        file_indentation = self.config_param.get(cfg.PROJECT_FILE_INDENTATION, cfg.PROJECT_FILE_INDENTATION_DEFAULT_VALUE)
        try:
            with open(file_name, "w") as f_out:
                f_out.write(json.dumps(pj, indent=file_indentation))
        except Exception:
            QMessageBox.critical(
                None,
                cfg.programName,
                "Error during file saving.",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )

    else:
        ethogram_data = tablib.Dataset()
        ethogram_data.title = "Ethogram"
        if self.leProjectName.text():
            ethogram_data.title = f"Ethogram of {self.leProjectName.text()} project"

        ethogram_data.headers = [
            "Behavior code",
            "Behavior type",
            "Description",
            "Key",
            "Color",
            "Behavioral category",
            "Excluded behaviors",
            "Modifiers",
            "Modifiers (JSON)",
        ]

        for r in range(self.twBehaviors.rowCount()):
            row: list = []
            for field in ("code", cfg.TYPE, "description", "key", cfg.COLOR, "category", "excluded"):
                row.append(self.twBehaviors.item(r, cfg.behavioursFields[field]).text())

            # modifiers
            if self.twBehaviors.item(r, cfg.behavioursFields[cfg.MODIFIERS]).text():
                # modifiers a string
                modifiers_dict = json.loads(self.twBehaviors.item(r, cfg.behavioursFields[cfg.MODIFIERS]).text())
                modifiers_list = []
                for key in modifiers_dict:
                    values = ",".join(modifiers_dict[key]["values"])
                    modifiers_list.append(f"{modifiers_dict[key]['name']}:{values}")
                row.append(";".join(modifiers_list))
                # modifiers as JSON
                row.append(self.twBehaviors.item(r, cfg.behavioursFields[cfg.MODIFIERS]).text())
            else:
                # modifiers a string
                row.append("")
                # modifiers as JSON
                row.append("")

            ethogram_data.append(row)

        ok, msg = export_observation.dataset_write(ethogram_data, file_name, output_format)
        if not ok:
            QMessageBox.critical(None, cfg.programName, msg, QMessageBox.Ok | QMessageBox.Default, QMessageBox.NoButton)


def export_subjects(self) -> None:
    """
    export the subjetcs list in various format
    """
    extended_file_formats: list = [
        cfg.TSV,
        cfg.CSV,
        cfg.ODS,
        cfg.XLSX,
        cfg.XLS,
        cfg.HTML,
    ]
    file_formats: list = [cfg.TSV_EXT, cfg.CSV_EXT, cfg.ODS_EXT, cfg.XLSX_EXT, cfg.XLS_EXT, cfg.HTML_EXT]

    filediag_func = QFileDialog().getSaveFileName

    file_name, filter_ = filediag_func(self, "Export the subjects list", "", ";;".join(extended_file_formats))
    if not file_name:
        return

    outputFormat = file_formats[extended_file_formats.index(filter_)]
    if pl.Path(file_name).suffix != "." + outputFormat:
        file_name = str(pl.Path(file_name)) + "." + outputFormat

    subjects_data = tablib.Dataset()
    subjects_data.title = "Subjects"
    if self.leProjectName.text():
        subjects_data.title = f"Subjects defined in the {self.leProjectName.text()} project"

    subjects_data.headers: list = [
        "Key",
        "Subject name",
        "Description",
    ]

    for r in range(self.twSubjects.rowCount()):
        row: list = []
        for idx, _ in enumerate(("Key", "Subject name", "Description")):
            row.append(self.twSubjects.item(r, idx).text())

        subjects_data.append(row)

    ok, msg = export_observation.dataset_write(subjects_data, file_name, outputFormat)
    if not ok:
        QMessageBox.critical(None, cfg.programName, msg, QMessageBox.Ok | QMessageBox.Default, QMessageBox.NoButton)


def select_behaviors(
    title: str = "Record value from external data file",
    text: str = "Behaviors",
    behavioral_categories: list = [],
    ethogram: dict = {},
    behavior_type=[cfg.STATE_EVENT, cfg.POINT_EVENT],
) -> list:
    """
    allow user to select behaviors to import

    Args:
        title (str): title of dialog box
        text (str): text of dialog box
        behavioral_categories (list): behavioral categories
        ethogram (dict): ethogram

    """

    paramPanelWindow = param_panel.Param_panel()
    paramPanelWindow.resize(800, 600)
    paramPanelWindow.setWindowTitle(title)
    paramPanelWindow.lbBehaviors.setText(text)
    for w in (
        paramPanelWindow.lwSubjects,
        paramPanelWindow.pbSelectAllSubjects,
        paramPanelWindow.pbUnselectAllSubjects,
        paramPanelWindow.pbReverseSubjectsSelection,
        paramPanelWindow.lbSubjects,
        paramPanelWindow.cbIncludeModifiers,
        paramPanelWindow.cbExcludeBehaviors,
        paramPanelWindow.frm_time,
        paramPanelWindow.frm_time_bin_size,
    ):
        w.setVisible(False)

    if behavioral_categories:
        categories = behavioral_categories
        # check if behavior not included in a category
        if "" in [ethogram[idx][cfg.BEHAVIOR_CATEGORY] for idx in ethogram if cfg.BEHAVIOR_CATEGORY in ethogram[idx]]:
            categories += [""]
    else:
        categories = ["###no category###"]

    for category in categories:
        if category != "###no category###":
            if category == "":
                paramPanelWindow.item = QListWidgetItem("No category")
                paramPanelWindow.item.setData(34, "No category")
            else:
                paramPanelWindow.item = QListWidgetItem(category)
                paramPanelWindow.item.setData(34, category)

            font = QFont()
            font.setBold(True)
            paramPanelWindow.item.setFont(font)
            paramPanelWindow.item.setData(33, "category")
            paramPanelWindow.item.setData(35, False)

            paramPanelWindow.lwBehaviors.addItem(paramPanelWindow.item)

        # check if behavior type must be shown
        for behavior in [ethogram[x][cfg.BEHAVIOR_CODE] for x in util.sorted_keys(ethogram)]:
            if (categories == ["###no category###"]) or (
                behavior
                in [
                    ethogram[x][cfg.BEHAVIOR_CODE]
                    for x in ethogram
                    if cfg.BEHAVIOR_CATEGORY in ethogram[x] and ethogram[x][cfg.BEHAVIOR_CATEGORY] == category
                ]
            ):
                paramPanelWindow.item = QListWidgetItem(behavior)
                paramPanelWindow.item.setCheckState(Qt.Unchecked)

                if category != "###no category###":
                    paramPanelWindow.item.setData(33, "behavior")
                    if category == "":
                        paramPanelWindow.item.setData(34, "No category")
                    else:
                        paramPanelWindow.item.setData(34, category)

                paramPanelWindow.lwBehaviors.addItem(paramPanelWindow.item)

    if paramPanelWindow.exec_():
        return paramPanelWindow.selectedBehaviors

    return []


def check_text_file_type(rows: list):
    """
    check text file
    returns separator and number of fields (if unique)
    """
    for separator in "\t,;":
        cs: list = []
        for row in rows:
            cs.append(row.count(separator))
        if len(set(cs)) == 1:
            return separator, cs[0] + 1
    return None, None


def import_ethogram_from_dict(self, project: dict):
    """
    Import behaviors from a BORIS project dictionary
    """
    # import behavioral_categories
    self.pj[cfg.BEHAVIORAL_CATEGORIES] = list(project.get(cfg.BEHAVIORAL_CATEGORIES, []))
    self.pj[cfg.BEHAVIORAL_CATEGORIES_CONF] = list(project.get(cfg.BEHAVIORAL_CATEGORIES_CONF, {}))

    # configuration of behaviours
    if not (cfg.ETHOGRAM in project and project[cfg.ETHOGRAM]):
        QMessageBox.warning(self, cfg.programName, "No behaviors configuration found in project")
        return

    if self.twBehaviors.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            ("Some behaviors are already configured. Do you want to append behaviors or replace them?"),
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )
        if response == cfg.REPLACE:
            self.twBehaviors.setRowCount(0)
            self.twBehaviors_cellChanged(0, 0)
        if response == cfg.CANCEL:
            return

    behaviors_to_import = select_behaviors(
        title="Select the behaviors to import",
        text="Behaviors",
        behavioral_categories=list(project.get(cfg.BEHAVIORAL_CATEGORIES, [])),
        ethogram=dict(project[cfg.ETHOGRAM]),
        behavior_type=[cfg.STATE_EVENT, cfg.POINT_EVENT],
    )

    for i in util.sorted_keys(project[cfg.ETHOGRAM]):
        if project[cfg.ETHOGRAM][i][cfg.BEHAVIOR_CODE] not in behaviors_to_import:
            continue

        self.twBehaviors.setRowCount(self.twBehaviors.rowCount() + 1)

        for field in project[cfg.ETHOGRAM][i]:
            item = QTableWidgetItem()

            if field == cfg.TYPE:
                item.setText(project[cfg.ETHOGRAM][i][field])
                item.setFlags(Qt.ItemIsEnabled)
                # item.setBackground(QColor(230, 230, 230))
                item.setBackground(self.not_editable_column_color())

            else:
                if field == cfg.MODIFIERS:
                    if isinstance(project[cfg.ETHOGRAM][i][field], str):
                        modif_set_dict = {}
                        if project[cfg.ETHOGRAM][i][field]:
                            modif_set_list = project[cfg.ETHOGRAM][i][field].split("|")
                            for modif_set in modif_set_list:
                                modif_set_dict[str(len(modif_set_dict))] = {
                                    "name": "",
                                    "type": cfg.SINGLE_SELECTION,
                                    "values": modif_set.split(","),
                                }
                        project[cfg.ETHOGRAM][i][field] = dict(modif_set_dict)
                    else:
                        item.setText(json.dumps(project[cfg.ETHOGRAM][i][field]))
                else:
                    item.setText(project[cfg.ETHOGRAM][i][field])

                if field not in cfg.ETHOGRAM_EDITABLE_FIELDS:
                    item.setFlags(Qt.ItemIsEnabled)
                    # item.setBackground(QColor(230, 230, 230))
                    item.setBackground(self.not_editable_column_color())

            self.twBehaviors.setItem(self.twBehaviors.rowCount() - 1, cfg.behavioursFields[field], item)

    self.twBehaviors.resizeColumnsToContents()


def load_dataframe_into_behaviors_tablewidget(self, df: pd.DataFrame) -> int:
    """
    Load pandas dataframe into the twBehaviors table widget

    Returns:
        int: 0 if no error else error code
    """

    expected_labels: list = [
        "Behavior code",
        "Behavior type",
        "Description",
        "Key",
        "Behavioral category",
        "Excluded behaviors",
    ]

    ethogram_header: dict = {
        "code": "Behavior code",
        "description": "Description",
        "key": "Key",
        "color": "Color",
        "category": "Behavioral category",
        "excluded": "Excluded behaviors",
        "modifiers": "modifiers (JSON)",
    }

    # change all column names to uppercase
    df.columns = df.columns.str.upper()

    for column in expected_labels:
        if column.upper() not in list(df.columns):
            QMessageBox.warning(
                None,
                cfg.programName,
                (
                    f"The {column} column was not found in the file header.<br>"
                    "For information the current file header contains the following labels:<br>"
                    f"{'<br>'.join(['<b>' + util.replace_leading_trailing_chars(x, ' ', '&#9608;') + '</b>' for x in df.columns])}<br>"
                    "<br>"
                    "The first row of the spreadsheet must contain the following labels:<br>"
                    f"{'<br>'.join(['<b>' + x + '</b>' for x in expected_labels])}<br>"
                    "<br>The order is not mandatory."
                ),
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return 1

    for _, row in df.iterrows():
        behavior = {"coding map": ""}
        for x in ethogram_header:
            if ethogram_header[x].upper() in row:
                behavior[x] = row[ethogram_header[x].upper()] if str(row[ethogram_header[x].upper()]) != "nan" else ""
            else:
                behavior[x] = ""

        self.twBehaviors.setRowCount(self.twBehaviors.rowCount() + 1)

        for field_type in cfg.behavioursFields:
            if field_type == cfg.TYPE:
                item = QTableWidgetItem(cfg.DEFAULT_BEHAVIOR_TYPE)
                # add type combobox
                if cfg.POINT in row["Behavior type".upper()].upper():
                    item = QTableWidgetItem(cfg.POINT_EVENT)
                elif cfg.STATE in row["Behavior type".upper()].upper():
                    item = QTableWidgetItem(cfg.STATE_EVENT)
                else:
                    QMessageBox.critical(
                        None,
                        cfg.programName,
                        f"{row['Behavior code']} has no behavior type (POINT or STATE)",
                        QMessageBox.Ok | QMessageBox.Default,
                        QMessageBox.NoButton,
                    )
                    return 2

            else:
                item = QTableWidgetItem(str(behavior[field_type]))

            if field_type not in cfg.ETHOGRAM_EDITABLE_FIELDS:
                item.setFlags(Qt.ItemIsEnabled)
                # item.setBackground(QColor(230, 230, 230))
                item.setBackground(self.not_editable_column_color())

            self.twBehaviors.setItem(self.twBehaviors.rowCount() - 1, cfg.behavioursFields[field_type], item)

    return 0


def import_behaviors_from_project(self):
    """
    import ethogram from a BORIS project file
    """
    file_name, _ = QFileDialog.getOpenFileName(
        self, "Import behaviors from BORIS project file", "", ("Project files (*.boris *.boris.gz);;All files (*)")
    )
    if not file_name:
        return
    _, _, project, _ = project_functions.open_project_json(file_name)

    import_ethogram_from_dict(self, project)


def import_behaviors_from_text_file(self):
    """
    Import ethogram from text file (CSV or TSV)
    """

    if self.twBehaviors.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            "There are behaviors already configured. Do you want to append behaviors or replace them?",
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )
        if response == cfg.CANCEL:
            return

    file_name, _ = QFileDialog.getOpenFileName(
        self, "Import behaviors from text file (CSV, TSV)", "", "Text files (*.txt *.tsv *.csv);;All files (*)"
    )

    if not file_name:
        return

    if pl.Path(file_name).suffix.upper() == ".CSV":
        delimiter = ","
    elif pl.Path(file_name).suffix.upper() == ".TSV":
        delimiter = "\t"
    else:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Comma Separated Values (,) or Tab Separated Values"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    try:
        df = pd.read_csv(file_name, delimiter=delimiter)
    except Exception:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Comma Separated Values (,) or Tab Separated Values"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    if self.twBehaviors.rowCount() and response == cfg.REPLACE:
        self.twBehaviors.setRowCount(0)

    load_dataframe_into_behaviors_tablewidget(self, df)


def import_behaviors_from_spreadsheet(self):
    """
    Import behaviors from a spreadsheet file (XLSX)
    """

    if self.twBehaviors.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            "There are behaviors already configured. Do you want to append behaviors or replace them?",
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )
        if response == cfg.CANCEL:
            return

    file_name, _ = QFileDialog.getOpenFileName(
        self, "Import behaviors from a spreadsheet file", "", "Spreadsheet files (*.xlsx *.ods);;All files (*)"
    )

    if not file_name:
        return

    if pl.Path(file_name).suffix.upper() == ".XLSX":
        engine = "openpyxl"
    elif pl.Path(file_name).suffix.upper() == ".ODS":
        engine = "odf"
    else:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Microsoft-Excel XLSX format or OpenDocument ODS"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    try:
        df = pd.read_excel(file_name, sheet_name=0, engine=engine)
    except Exception:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Microsoft-Excel XLSX format or OpenDocument ODS"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    if self.twBehaviors.rowCount() and response == cfg.REPLACE:
        self.twBehaviors.setRowCount(0)

    load_dataframe_into_behaviors_tablewidget(self, df)


def import_behaviors_from_clipboard(self):
    """
    import ethogram from clipboard
    """

    cb = QApplication.clipboard()
    cb_text = cb.text()
    if not cb_text:
        QMessageBox.warning(
            None,
            cfg.programName,
            "The clipboard is empty",
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    if self.twBehaviors.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            "Some behaviors are already configured. Do you want to append behaviors or replace them?",
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )
        if response == cfg.CANCEL:
            return

        if response == cfg.REPLACE:
            self.twBehaviors.setRowCount(0)

    cb_text_splitted = cb_text.split("\n")
    while "" in cb_text_splitted:
        cb_text_splitted.remove("")

    if len(set([len(x.split("\t")) for x in cb_text_splitted])) != 1:
        QMessageBox.warning(
            None,
            cfg.programName,
            (
                "The clipboard content does not have a constant number of fields.<br>"
                "From your spreadsheet: CTRL + A (select all cells), CTRL + C (copy to clipboard)"
            ),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    for row in cb_text_splitted:
        if set(row.split("\t")) != set([""]):
            behavior = {"type": cfg.DEFAULT_BEHAVIOR_TYPE}
            for idx, field in enumerate(row.split("\t")):
                if idx == 0:
                    behavior["type"] = (
                        cfg.STATE_EVENT if cfg.STATE in field.upper() else (cfg.POINT_EVENT if cfg.POINT in field.upper() else "")
                    )
                if idx == 1:
                    behavior["key"] = field.strip() if len(field.strip()) == 1 else ""
                if idx == 2:
                    behavior["code"] = field.strip()
                if idx == 3:
                    behavior["description"] = field.strip()
                if idx == 4:
                    behavior["category"] = field.strip()

            self.twBehaviors.setRowCount(self.twBehaviors.rowCount() + 1)

            for field_type in cfg.behavioursFields:
                if field_type == cfg.TYPE:
                    item = QTableWidgetItem(behavior.get("type", cfg.DEFAULT_BEHAVIOR_TYPE))
                else:
                    item = QTableWidgetItem(behavior.get(field_type, ""))

                if field_type not in cfg.ETHOGRAM_EDITABLE_FIELDS:  # [TYPE, "excluded", "coding map", "modifiers", "category"]:
                    item.setFlags(Qt.ItemIsEnabled)
                    # item.setBackground(QColor(230, 230, 230))
                    item.setBackground(self.not_editable_column_color())

                self.twBehaviors.setItem(self.twBehaviors.rowCount() - 1, cfg.behavioursFields[field_type], item)


def import_behaviors_from_JWatcher(self):
    """
    import behaviors configuration from JWatcher (GDF file)
    """

    if self.twBehaviors.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            "There are behaviors already configured. Do you want to append behaviors or replace them?",
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )
        if response == cfg.CANCEL:
            return

    fileName, _ = QFileDialog().getOpenFileName(self, "Import behaviors from JWatcher", "", "Global Definition File (*.gdf);;All files (*)")

    if not fileName:
        return
    if self.twBehaviors.rowCount() and response == cfg.REPLACE:
        self.twBehaviors.setRowCount(0)

    with open(fileName, "r") as f:
        rows = f.readlines()

    for idx, row in enumerate(rows):
        if row and row[0] == "#":
            continue

        if "Behavior.name." in row and "=" in row:
            key, code = row.split("=")
            key = key.replace("Behavior.name.", "")
            # read description
            if idx < len(rows) and "Behavior.description." in rows[idx + 1]:
                description = rows[idx + 1].split("=")[-1]

            behavior = {
                "key": key,
                "code": code,
                "description": description,
                "modifiers": "",
                "excluded": "",
                "coding map": "",
                "category": "",
            }

            self.twBehaviors.setRowCount(self.twBehaviors.rowCount() + 1)

            for field_type in cfg.behavioursFields:
                if field_type == cfg.TYPE:
                    item = QTableWidgetItem(cfg.DEFAULT_BEHAVIOR_TYPE)
                else:
                    item = QTableWidgetItem(behavior[field_type])

                if field_type in [cfg.TYPE, "excluded", "category", "coding map", "modifiers"]:
                    item.setFlags(Qt.ItemIsEnabled)
                    # item.setBackground(QColor(230, 230, 230))
                    item.setBackground(self.not_editable_column_color())

                self.twBehaviors.setItem(self.twBehaviors.rowCount() - 1, cfg.behavioursFields[field_type], item)


def import_behaviors_from_repository(self):
    """
    import behaviors from the BORIS ethogram repository
    """

    try:
        ethogram_list = urllib.request.urlopen(f"{cfg.ETHOGRAM_REPOSITORY_URL}/ethogram_list.json").read().strip().decode("utf-8")
    except Exception:
        QMessageBox.critical(self, cfg.programName, "An error occured during retrieving the ethogram list from BORIS repository")
        return

    try:
        ethogram_list_list = json.loads(ethogram_list)
    except Exception:
        QMessageBox.critical(self, cfg.programName, "An error occured during loading ethogram list from BORIS repository")
        return

    choice_dialog = dialog.ChooseObservationsToImport(
        "Choose the ethogram to import:", sorted([f"{x['species']} by {x['author']}" for x in ethogram_list_list])
    )
    while True:
        if not choice_dialog.exec_():
            return

        if len(choice_dialog.get_selected_observations()) == 0:
            QMessageBox.critical(self, cfg.programName, "Choose one ethogram")
            continue

        if len(choice_dialog.get_selected_observations()) > 1:
            QMessageBox.critical(self, cfg.programName, "Choose only one ethogram")
            continue

        break

    for x in ethogram_list_list:
        if f"{x['species']} by {x['author']}" == choice_dialog.get_selected_observations()[0]:
            file_name = x["file name"]
            break

    try:
        boris_project_str = urllib.request.urlopen(f"{cfg.ETHOGRAM_REPOSITORY_URL}/{file_name}").read().strip().decode("utf-8")
    except Exception:
        QMessageBox.critical(self, cfg.programName, f"An error occured during retrieving {file_name} from BORIS repository")
        return
    boris_project = json.loads(boris_project_str)

    import_ethogram_from_dict(self, boris_project)


def load_dataframe_into_subjects_tablewidget(self, df: pd.DataFrame) -> int:
    """
    Load pandas dataframe into the twSubjects table widget

    Returns:
        int: 0 if no error else error code

    """

    expected_labels: list = ["Key", "Subject name", "Description"]

    # change all column names to uppercase
    df.columns = df.columns.str.upper()

    for column in expected_labels:
        if column.upper() not in list(df.columns):
            QMessageBox.warning(
                None,
                cfg.programName,
                (
                    f"The column {column} was not found in the file header.<br>"
                    "The first row of spreadsheet must contain the following labels:<br>"
                    "Subject name, Description, Key<br>"
                    "The order is not mandatory."
                ),
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return 1

    for _, row in df.iterrows():
        self.twSubjects.setRowCount(self.twSubjects.rowCount() + 1)

        for idx, field in enumerate(expected_labels):
            self.twSubjects.setItem(
                self.twSubjects.rowCount() - 1,
                idx,
                QTableWidgetItem(str(row[field.upper()]) if str(row[field.upper()]) != "nan" else ""),
            )

    return 0


def import_subjects_from_clipboard(self):
    """
    import subjects from clipboard
    """
    cb = QApplication.clipboard()
    cb_text = cb.text()
    if not cb_text:
        QMessageBox.warning(
            None,
            cfg.programName,
            "The clipboard is empty",
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    if self.twSubjects.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            "Some subjects are already configured. Do you want to append subjects or replace them?",
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )
        if response == cfg.CANCEL:
            return

        if response == cfg.REPLACE:
            self.twSubjects.setRowCount(0)

    cb_text_splitted = cb_text.split("\n")

    if len(set([len(x.split("\t")) for x in cb_text_splitted])) != 1:
        QMessageBox.warning(
            None,
            cfg.programName,
            (
                "The clipboard content does not have a constant number of fields.<br>"
                "From your spreadsheet: CTRL + A (select all cells), CTRL + C (copy to clipboard)"
            ),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    for row in cb_text_splitted:
        if set(row.split("\t")) != set([""]):
            subject = {}
            for idx, field in enumerate(row.split("\t")):
                if idx == 0:
                    subject["key"] = field.strip() if len(field.strip()) == 1 else ""
                if idx == 1:
                    subject[cfg.SUBJECT_NAME] = field.strip()
                if idx == 2:
                    subject["description"] = field.strip()

            self.twSubjects.setRowCount(self.twSubjects.rowCount() + 1)

            for idx, field_name in enumerate(cfg.subjectsFields):
                item = QTableWidgetItem(subject.get(field_name, ""))
                self.twSubjects.setItem(self.twSubjects.rowCount() - 1, idx, item)


def import_subjects_from_project(self):
    """
    import subjects from a BORIS project
    """

    file_name, _ = QFileDialog().getOpenFileName(
        self, "Import subjects from project file", "", ("Project files (*.boris *.boris.gz);;All files (*)")
    )
    if not file_name:
        return

    _, _, project, _ = project_functions.open_project_json(file_name)

    if "error" in project:
        logging.debug(project["error"])
        QMessageBox.critical(self, cfg.programName, project["error"])
        return

    # configuration of subjects
    if not (cfg.SUBJECTS in project and project[cfg.SUBJECTS]):
        QMessageBox.warning(self, cfg.programName, "No subjects configuration found in project")
        return

    if self.twSubjects.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            ("There are subjects already configured. Do you want to append subjects or replace them?"),
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )

        if response == cfg.REPLACE:
            self.twSubjects.setRowCount(0)

        if response == cfg.CANCEL:
            return

    for idx in util.sorted_keys(project[cfg.SUBJECTS]):
        self.twSubjects.setRowCount(self.twSubjects.rowCount() + 1)

        for idx2, sbjField in enumerate(cfg.subjectsFields):
            if sbjField in project[cfg.SUBJECTS][idx]:
                self.twSubjects.setItem(
                    self.twSubjects.rowCount() - 1,
                    idx2,
                    QTableWidgetItem(project[cfg.SUBJECTS][idx][sbjField]),
                )
            else:
                self.twSubjects.setItem(self.twSubjects.rowCount() - 1, idx2, QTableWidgetItem(""))

    self.twSubjects.resizeColumnsToContents()


def import_subjects_from_text_file(self):
    """
    import subjects from a text file (CSV or TSV)
    """

    if self.twSubjects.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            ("There are subjects already configured. Do you want to append subjects or replace them?"),
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )

        if response == cfg.CANCEL:
            return

    file_name, _ = QFileDialog().getOpenFileName(
        self, "Import behaviors from text file (CSV, TSV)", "", "Text files (*.txt *.tsv *.csv);;All files (*)"
    )
    if not file_name:
        return

    if self.twSubjects.rowCount() and response == cfg.REPLACE:
        self.twSubjects.setRowCount(0)

    if pl.Path(file_name).suffix.upper() == ".CSV":
        delimiter = ","
    elif pl.Path(file_name).suffix.upper() == ".TSV":
        delimiter = "\t"
    else:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Comma Separated Values (,) or Tab Separated Values"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    try:
        df = pd.read_csv(file_name, delimiter=delimiter)
    except Exception:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Comma Separated Values (,) or Tab Separated Values"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    load_dataframe_into_subjects_tablewidget(self, df)


def import_subjects_from_spreadsheet(self):
    """
    import subjects from a spreadsheet file (XLSX or ODS)
    """

    if self.twSubjects.rowCount():
        response = dialog.MessageDialog(
            cfg.programName,
            ("There are subjects already configured. Do you want to append subjects or replace them?"),
            [cfg.APPEND, cfg.REPLACE, cfg.CANCEL],
        )

        if response == cfg.CANCEL:
            return

    file_name, _ = QFileDialog().getOpenFileName(
        self, "Import subjects from a spreadsheet file", "", "Spreadsheet files (*.xlsx *.ods);;All files (*)"
    )
    if not file_name:
        return

    if self.twSubjects.rowCount() and response == cfg.REPLACE:
        self.twSubjects.setRowCount(0)

    if pl.Path(file_name).suffix.upper() == ".XLSX":
        engine = "openpyxl"
    elif pl.Path(file_name).suffix.upper() == ".ODS":
        engine = "odf"
    else:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Microsoft-Excel XLSX format or OpenDocument ODS"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    try:
        df = pd.read_excel(file_name, sheet_name=0, engine=engine)
    except Exception:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Microsoft-Excel XLSX format or OpenDocument ODS"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    load_dataframe_into_subjects_tablewidget(self, df)


def import_indep_variables_from_project(self):
    """
    import independent variables from another project
    """

    file_name, _ = QFileDialog().getOpenFileName(
        self,
        "Import independent variables from project file",
        "",
        ("Project files (*.boris *.boris.gz);;All files (*)"),
    )
    if not file_name:
        return

    _, _, project, _ = project_functions.open_project_json(file_name)

    if "error" in project:
        logging.debug(project["error"])
        QMessageBox.critical(self, cfg.programName, project["error"])
        return

    # independent variables
    if not (cfg.INDEPENDENT_VARIABLES in project and project[cfg.INDEPENDENT_VARIABLES]):
        QMessageBox.warning(self, cfg.programName, "No independent variables found in project")
        return

    # check if variables are already present
    existing_var = []

    for r in range(self.twVariables.rowCount()):
        existing_var.append(self.twVariables.item(r, 0).text().strip().upper())

    for i in util.sorted_keys(project[cfg.INDEPENDENT_VARIABLES]):
        self.twVariables.setRowCount(self.twVariables.rowCount() + 1)
        flag_renamed = False
        for idx, field in enumerate(cfg.tw_indVarFields):
            item = QTableWidgetItem()
            if field in project[cfg.INDEPENDENT_VARIABLES][i]:
                if field == "label":
                    txt = project[cfg.INDEPENDENT_VARIABLES][i]["label"].strip()
                    while txt.upper() in existing_var:
                        txt += "_2"
                        flag_renamed = True
                else:
                    txt = project[cfg.INDEPENDENT_VARIABLES][i][field].strip()
                item.setText(txt)
            else:
                item.setText("")
            self.twVariables.setItem(self.twVariables.rowCount() - 1, idx, item)

    self.twVariables.resizeColumnsToContents()
    if flag_renamed:
        QMessageBox.information(self, cfg.programName, "Some variables already present were renamed")



import enum


class Bound(enum.Enum):
    """
    Bound types, either CLOSED for inclusive, or OPEN for exclusive.
    """
    CLOSED = True
    OPEN = False

    def __bool__(self):
        raise ValueError('The truth value of a bound is ambiguous.')

    def __invert__(self):
        return Bound.CLOSED if self is Bound.OPEN else Bound.OPEN

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class _Singleton():
    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super(_Singleton, cls).__new__(cls)
        return cls.__instance


class _PInf(_Singleton):
    """
    Represent positive infinity.
    """

    def __neg__(self): return _NInf()

    def __lt__(self, o): return False

    def __le__(self, o): return isinstance(o, _PInf)

    def __gt__(self, o): return not isinstance(o, _PInf)

    def __ge__(self, o): return True

    def __eq__(self, o): return isinstance(o, _PInf)

    def __repr__(self): return '+inf'

    def __hash__(self): return hash(float('+inf'))


class _NInf(_Singleton):
    """
    Represent negative infinity.
    """

    def __neg__(self): return _PInf()

    def __lt__(self, o): return not isinstance(o, _NInf)

    def __le__(self, o): return True

    def __gt__(self, o): return False

    def __ge__(self, o): return isinstance(o, _NInf)

    def __eq__(self, o): return isinstance(o, _NInf)

    def __repr__(self): return '-inf'

    def __hash__(self): return hash(float('-inf'))


# Positive infinity
inf = _PInf()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'observation.ui'
##
## Created by: Qt User Interface Compiler version 6.8.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QDateTimeEdit,
    QDoubleSpinBox, QGroupBox, QHBoxLayout, QHeaderView,
    QLabel, QLineEdit, QListWidget, QListWidgetItem,
    QPushButton, QRadioButton, QSizePolicy, QSpacerItem,
    QSpinBox, QSplitter, QStackedWidget, QTabWidget,
    QTableWidget, QTableWidgetItem, QTextEdit, QVBoxLayout,
    QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(1278, 677)
        self.verticalLayout_6 = QVBoxLayout(Form)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label = QLabel(Form)
        self.label.setObjectName(u"label")

        self.horizontalLayout_2.addWidget(self.label)

        self.lb_star = QLabel(Form)
        self.lb_star.setObjectName(u"lb_star")
        font = QFont()
        font.setPointSize(14)
        self.lb_star.setFont(font)
        self.lb_star.setStyleSheet(u"color: red")

        self.horizontalLayout_2.addWidget(self.lb_star)

        self.leObservationId = QLineEdit(Form)
        self.leObservationId.setObjectName(u"leObservationId")

        self.horizontalLayout_2.addWidget(self.leObservationId)

        self.label_8 = QLabel(Form)
        self.label_8.setObjectName(u"label_8")

        self.horizontalLayout_2.addWidget(self.label_8)

        self.dteDate = QDateTimeEdit(Form)
        self.dteDate.setObjectName(u"dteDate")
        self.dteDate.setCalendarPopup(True)

        self.horizontalLayout_2.addWidget(self.dteDate)


        self.verticalLayout_6.addLayout(self.horizontalLayout_2)

        self.splitter = QSplitter(Form)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout_2 = QVBoxLayout(self.layoutWidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.label_9 = QLabel(self.layoutWidget)
        self.label_9.setObjectName(u"label_9")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_9.sizePolicy().hasHeightForWidth())
        self.label_9.setSizePolicy(sizePolicy)
        self.label_9.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)

        self.verticalLayout_2.addWidget(self.label_9)

        self.teDescription = QTextEdit(self.layoutWidget)
        self.teDescription.setObjectName(u"teDescription")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.teDescription.sizePolicy().hasHeightForWidth())
        self.teDescription.setSizePolicy(sizePolicy1)
        self.teDescription.setMaximumSize(QSize(16777215, 16777215))
        self.teDescription.setAcceptDrops(False)

        self.verticalLayout_2.addWidget(self.teDescription)

        self.cb_time_offset = QCheckBox(self.layoutWidget)
        self.cb_time_offset.setObjectName(u"cb_time_offset")

        self.verticalLayout_2.addWidget(self.cb_time_offset)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.lbTimeOffset = QLabel(self.layoutWidget)
        self.lbTimeOffset.setObjectName(u"lbTimeOffset")

        self.horizontalLayout_6.addWidget(self.lbTimeOffset)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_2)


        self.verticalLayout_2.addLayout(self.horizontalLayout_6)

        self.cb_observation_time_interval = QCheckBox(self.layoutWidget)
        self.cb_observation_time_interval.setObjectName(u"cb_observation_time_interval")

        self.verticalLayout_2.addWidget(self.cb_observation_time_interval)

        self.splitter.addWidget(self.layoutWidget)
        self.layoutWidget1 = QWidget(self.splitter)
        self.layoutWidget1.setObjectName(u"layoutWidget1")
        self.verticalLayout_11 = QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_11.setObjectName(u"verticalLayout_11")
        self.verticalLayout_11.setContentsMargins(0, 0, 0, 0)
        self.label_3 = QLabel(self.layoutWidget1)
        self.label_3.setObjectName(u"label_3")

        self.verticalLayout_11.addWidget(self.label_3)

        self.twIndepVariables = QTableWidget(self.layoutWidget1)
        if (self.twIndepVariables.columnCount() < 3):
            self.twIndepVariables.setColumnCount(3)
        __qtablewidgetitem = QTableWidgetItem()
        self.twIndepVariables.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.twIndepVariables.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.twIndepVariables.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        self.twIndepVariables.setObjectName(u"twIndepVariables")

        self.verticalLayout_11.addWidget(self.twIndepVariables)

        self.splitter.addWidget(self.layoutWidget1)

        self.verticalLayout_6.addWidget(self.splitter)

        self.gb_observation_type = QGroupBox(Form)
        self.gb_observation_type.setObjectName(u"gb_observation_type")
        self.horizontalLayout_7 = QHBoxLayout(self.gb_observation_type)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.rb_media_files = QRadioButton(self.gb_observation_type)
        self.rb_media_files.setObjectName(u"rb_media_files")

        self.horizontalLayout_4.addWidget(self.rb_media_files)

        self.rb_live = QRadioButton(self.gb_observation_type)
        self.rb_live.setObjectName(u"rb_live")

        self.horizontalLayout_4.addWidget(self.rb_live)

        self.rb_images = QRadioButton(self.gb_observation_type)
        self.rb_images.setObjectName(u"rb_images")
        self.rb_images.setEnabled(True)

        self.horizontalLayout_4.addWidget(self.rb_images)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_4)


        self.horizontalLayout_7.addLayout(self.horizontalLayout_4)


        self.verticalLayout_6.addWidget(self.gb_observation_type)

        self.sw_observation_type = QStackedWidget(Form)
        self.sw_observation_type.setObjectName(u"sw_observation_type")
        self.page = QWidget()
        self.page.setObjectName(u"page")
        self.sw_observation_type.addWidget(self.page)
        self.pg_media_files = QWidget()
        self.pg_media_files.setObjectName(u"pg_media_files")
        self.verticalLayout_7 = QVBoxLayout(self.pg_media_files)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.tabWidget = QTabWidget(self.pg_media_files)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tab_player_1 = QWidget()
        self.tab_player_1.setObjectName(u"tab_player_1")
        self.verticalLayout = QVBoxLayout(self.tab_player_1)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.twVideo1 = QTableWidget(self.tab_player_1)
        if (self.twVideo1.columnCount() < 7):
            self.twVideo1.setColumnCount(7)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(0, __qtablewidgetitem3)
        __qtablewidgetitem4 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(1, __qtablewidgetitem4)
        __qtablewidgetitem5 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(2, __qtablewidgetitem5)
        __qtablewidgetitem6 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(3, __qtablewidgetitem6)
        __qtablewidgetitem7 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(4, __qtablewidgetitem7)
        __qtablewidgetitem8 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(5, __qtablewidgetitem8)
        __qtablewidgetitem9 = QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(6, __qtablewidgetitem9)
        self.twVideo1.setObjectName(u"twVideo1")
        self.twVideo1.setEditTriggers(QAbstractItemView.EditTrigger.AnyKeyPressed|QAbstractItemView.EditTrigger.DoubleClicked|QAbstractItemView.EditTrigger.EditKeyPressed)
        self.twVideo1.setAlternatingRowColors(True)
        self.twVideo1.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
        self.twVideo1.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.twVideo1.setTextElideMode(Qt.TextElideMode.ElideNone)

        self.verticalLayout_3.addWidget(self.twVideo1)


        self.verticalLayout.addLayout(self.verticalLayout_3)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.pbAddVideo = QPushButton(self.tab_player_1)
        self.pbAddVideo.setObjectName(u"pbAddVideo")

        self.horizontalLayout_3.addWidget(self.pbAddVideo)

        self.pbRemoveVideo = QPushButton(self.tab_player_1)
        self.pbRemoveVideo.setObjectName(u"pbRemoveVideo")

        self.horizontalLayout_3.addWidget(self.pbRemoveVideo)

        self.pb_use_media_file_name_as_obsid = QPushButton(self.tab_player_1)
        self.pb_use_media_file_name_as_obsid.setObjectName(u"pb_use_media_file_name_as_obsid")

        self.horizontalLayout_3.addWidget(self.pb_use_media_file_name_as_obsid)

        self.horizontalSpacer_6 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_6)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_15 = QHBoxLayout()
        self.horizontalLayout_15.setObjectName(u"horizontalLayout_15")
        self.cbVisualizeSpectrogram = QCheckBox(self.tab_player_1)
        self.cbVisualizeSpectrogram.setObjectName(u"cbVisualizeSpectrogram")

        self.horizontalLayout_15.addWidget(self.cbVisualizeSpectrogram)

        self.cb_visualize_waveform = QCheckBox(self.tab_player_1)
        self.cb_visualize_waveform.setObjectName(u"cb_visualize_waveform")

        self.horizontalLayout_15.addWidget(self.cb_visualize_waveform)

        self.cb_media_creation_date_as_offset = QCheckBox(self.tab_player_1)
        self.cb_media_creation_date_as_offset.setObjectName(u"cb_media_creation_date_as_offset")
        self.cb_media_creation_date_as_offset.setEnabled(False)

        self.horizontalLayout_15.addWidget(self.cb_media_creation_date_as_offset)

        self.horizontalSpacer_11 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_15.addItem(self.horizontalSpacer_11)


        self.verticalLayout.addLayout(self.horizontalLayout_15)

        self.horizontalLayout_12 = QHBoxLayout()
        self.horizontalLayout_12.setObjectName(u"horizontalLayout_12")
        self.label_5 = QLabel(self.tab_player_1)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_12.addWidget(self.label_5)

        self.sb_media_scan_sampling = QSpinBox(self.tab_player_1)
        self.sb_media_scan_sampling.setObjectName(u"sb_media_scan_sampling")
        self.sb_media_scan_sampling.setMaximum(1000000)

        self.horizontalLayout_12.addWidget(self.sb_media_scan_sampling)

        self.label_2 = QLabel(self.tab_player_1)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout_12.addWidget(self.label_2)

        self.sb_image_display_duration = QSpinBox(self.tab_player_1)
        self.sb_image_display_duration.setObjectName(u"sb_image_display_duration")
        self.sb_image_display_duration.setMinimum(1)
        self.sb_image_display_duration.setMaximum(86400)

        self.horizontalLayout_12.addWidget(self.sb_image_display_duration)

        self.horizontalSpacer_8 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_12.addItem(self.horizontalSpacer_8)


        self.verticalLayout.addLayout(self.horizontalLayout_12)

        self.cbCloseCurrentBehaviorsBetweenVideo = QCheckBox(self.tab_player_1)
        self.cbCloseCurrentBehaviorsBetweenVideo.setObjectName(u"cbCloseCurrentBehaviorsBetweenVideo")
        self.cbCloseCurrentBehaviorsBetweenVideo.setEnabled(False)

        self.verticalLayout.addWidget(self.cbCloseCurrentBehaviorsBetweenVideo)

        self.tabWidget.addTab(self.tab_player_1, "")
        self.tab_data_files = QWidget()
        self.tab_data_files.setObjectName(u"tab_data_files")
        self.verticalLayout_17 = QVBoxLayout(self.tab_data_files)
        self.verticalLayout_17.setObjectName(u"verticalLayout_17")
        self.splitter_5 = QSplitter(self.tab_data_files)
        self.splitter_5.setObjectName(u"splitter_5")
        self.splitter_5.setOrientation(Qt.Orientation.Vertical)
        self.layoutWidget_4 = QWidget(self.splitter_5)
        self.layoutWidget_4.setObjectName(u"layoutWidget_4")
        self.verticalLayout_15 = QVBoxLayout(self.layoutWidget_4)
        self.verticalLayout_15.setObjectName(u"verticalLayout_15")
        self.verticalLayout_15.setContentsMargins(0, 0, 0, 0)
        self.label_7 = QLabel(self.layoutWidget_4)
        self.label_7.setObjectName(u"label_7")

        self.verticalLayout_15.addWidget(self.label_7)

        self.tw_data_files = QTableWidget(self.layoutWidget_4)
        if (self.tw_data_files.columnCount() < 9):
            self.tw_data_files.setColumnCount(9)
        __qtablewidgetitem10 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(0, __qtablewidgetitem10)
        __qtablewidgetitem11 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(1, __qtablewidgetitem11)
        __qtablewidgetitem12 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(2, __qtablewidgetitem12)
        __qtablewidgetitem13 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(3, __qtablewidgetitem13)
        __qtablewidgetitem14 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(4, __qtablewidgetitem14)
        __qtablewidgetitem15 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(5, __qtablewidgetitem15)
        __qtablewidgetitem16 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(6, __qtablewidgetitem16)
        __qtablewidgetitem17 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(7, __qtablewidgetitem17)
        __qtablewidgetitem18 = QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(8, __qtablewidgetitem18)
        self.tw_data_files.setObjectName(u"tw_data_files")
        self.tw_data_files.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)

        self.verticalLayout_15.addWidget(self.tw_data_files)

        self.splitter_5.addWidget(self.layoutWidget_4)
        self.layoutWidget_5 = QWidget(self.splitter_5)
        self.layoutWidget_5.setObjectName(u"layoutWidget_5")
        self.verticalLayout_16 = QVBoxLayout(self.layoutWidget_5)
        self.verticalLayout_16.setObjectName(u"verticalLayout_16")
        self.verticalLayout_16.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.pb_add_data_file = QPushButton(self.layoutWidget_5)
        self.pb_add_data_file.setObjectName(u"pb_add_data_file")

        self.horizontalLayout_5.addWidget(self.pb_add_data_file)

        self.pb_remove_data_file = QPushButton(self.layoutWidget_5)
        self.pb_remove_data_file.setObjectName(u"pb_remove_data_file")

        self.horizontalLayout_5.addWidget(self.pb_remove_data_file)

        self.pb_view_data_head = QPushButton(self.layoutWidget_5)
        self.pb_view_data_head.setObjectName(u"pb_view_data_head")

        self.horizontalLayout_5.addWidget(self.pb_view_data_head)

        self.pb_plot_data = QPushButton(self.layoutWidget_5)
        self.pb_plot_data.setObjectName(u"pb_plot_data")

        self.horizontalLayout_5.addWidget(self.pb_plot_data)

        self.horizontalSpacer_7 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_7)


        self.verticalLayout_16.addLayout(self.horizontalLayout_5)

        self.verticalSpacer_4 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_16.addItem(self.verticalSpacer_4)

        self.splitter_5.addWidget(self.layoutWidget_5)

        self.verticalLayout_17.addWidget(self.splitter_5)

        self.tabWidget.addTab(self.tab_data_files, "")

        self.verticalLayout_7.addWidget(self.tabWidget)

        self.sw_observation_type.addWidget(self.pg_media_files)
        self.pg_live = QWidget()
        self.pg_live.setObjectName(u"pg_live")
        self.verticalLayout_4 = QVBoxLayout(self.pg_live)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_8 = QVBoxLayout()
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.horizontalLayout_8 = QHBoxLayout()
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.label_4 = QLabel(self.pg_live)
        self.label_4.setObjectName(u"label_4")

        self.horizontalLayout_8.addWidget(self.label_4)

        self.sbScanSampling = QSpinBox(self.pg_live)
        self.sbScanSampling.setObjectName(u"sbScanSampling")
        self.sbScanSampling.setMaximum(1000000)

        self.horizontalLayout_8.addWidget(self.sbScanSampling)

        self.label_6 = QLabel(self.pg_live)
        self.label_6.setObjectName(u"label_6")

        self.horizontalLayout_8.addWidget(self.label_6)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer)


        self.verticalLayout_8.addLayout(self.horizontalLayout_8)

        self.cb_start_from_current_time = QCheckBox(self.pg_live)
        self.cb_start_from_current_time.setObjectName(u"cb_start_from_current_time")

        self.verticalLayout_8.addWidget(self.cb_start_from_current_time)

        self.rb_day_time = QRadioButton(self.pg_live)
        self.rb_day_time.setObjectName(u"rb_day_time")
        self.rb_day_time.setEnabled(False)
        self.rb_day_time.setChecked(True)

        self.verticalLayout_8.addWidget(self.rb_day_time)

        self.rb_epoch_time = QRadioButton(self.pg_live)
        self.rb_epoch_time.setObjectName(u"rb_epoch_time")
        self.rb_epoch_time.setEnabled(False)

        self.verticalLayout_8.addWidget(self.rb_epoch_time)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_8.addItem(self.verticalSpacer_2)


        self.verticalLayout_4.addLayout(self.verticalLayout_8)

        self.sw_observation_type.addWidget(self.pg_live)
        self.pg_images = QWidget()
        self.pg_images.setObjectName(u"pg_images")
        self.verticalLayout_5 = QVBoxLayout(self.pg_images)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.horizontalLayout_9 = QHBoxLayout()
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.lw_images_directory = QListWidget(self.pg_images)
        self.lw_images_directory.setObjectName(u"lw_images_directory")

        self.horizontalLayout_9.addWidget(self.lw_images_directory)


        self.verticalLayout_5.addLayout(self.horizontalLayout_9)

        self.lb_images_info = QLabel(self.pg_images)
        self.lb_images_info.setObjectName(u"lb_images_info")

        self.verticalLayout_5.addWidget(self.lb_images_info)

        self.horizontalLayout_14 = QHBoxLayout()
        self.horizontalLayout_14.setObjectName(u"horizontalLayout_14")
        self.pb_add_directory = QPushButton(self.pg_images)
        self.pb_add_directory.setObjectName(u"pb_add_directory")

        self.horizontalLayout_14.addWidget(self.pb_add_directory)

        self.pb_remove_directory = QPushButton(self.pg_images)
        self.pb_remove_directory.setObjectName(u"pb_remove_directory")

        self.horizontalLayout_14.addWidget(self.pb_remove_directory)

        self.horizontalSpacer_10 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_14.addItem(self.horizontalSpacer_10)


        self.verticalLayout_5.addLayout(self.horizontalLayout_14)

        self.horizontalLayout_13 = QHBoxLayout()
        self.horizontalLayout_13.setObjectName(u"horizontalLayout_13")
        self.pb_use_img_dir_as_obsid = QPushButton(self.pg_images)
        self.pb_use_img_dir_as_obsid.setObjectName(u"pb_use_img_dir_as_obsid")

        self.horizontalLayout_13.addWidget(self.pb_use_img_dir_as_obsid)

        self.horizontalSpacer_9 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_13.addItem(self.horizontalSpacer_9)


        self.verticalLayout_5.addLayout(self.horizontalLayout_13)

        self.groupBox = QGroupBox(self.pg_images)
        self.groupBox.setObjectName(u"groupBox")
        self.horizontalLayout_11 = QHBoxLayout(self.groupBox)
        self.horizontalLayout_11.setObjectName(u"horizontalLayout_11")
        self.horizontalLayout_10 = QHBoxLayout()
        self.horizontalLayout_10.setObjectName(u"horizontalLayout_10")
        self.rb_no_time = QRadioButton(self.groupBox)
        self.rb_no_time.setObjectName(u"rb_no_time")

        self.horizontalLayout_10.addWidget(self.rb_no_time)

        self.rb_use_exif = QRadioButton(self.groupBox)
        self.rb_use_exif.setObjectName(u"rb_use_exif")

        self.horizontalLayout_10.addWidget(self.rb_use_exif)

        self.rb_time_lapse = QRadioButton(self.groupBox)
        self.rb_time_lapse.setObjectName(u"rb_time_lapse")

        self.horizontalLayout_10.addWidget(self.rb_time_lapse)

        self.sb_time_lapse = QDoubleSpinBox(self.groupBox)
        self.sb_time_lapse.setObjectName(u"sb_time_lapse")
        self.sb_time_lapse.setDecimals(3)
        self.sb_time_lapse.setMaximum(86400.000000000000000)

        self.horizontalLayout_10.addWidget(self.sb_time_lapse)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_10.addItem(self.horizontalSpacer_5)


        self.horizontalLayout_11.addLayout(self.horizontalLayout_10)


        self.verticalLayout_5.addWidget(self.groupBox)

        self.sw_observation_type.addWidget(self.pg_images)

        self.verticalLayout_6.addWidget(self.sw_observation_type)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_3)

        self.pbCancel = QPushButton(Form)
        self.pbCancel.setObjectName(u"pbCancel")

        self.horizontalLayout.addWidget(self.pbCancel)

        self.pbSave = QPushButton(Form)
        self.pbSave.setObjectName(u"pbSave")

        self.horizontalLayout.addWidget(self.pbSave)

        self.pbLaunch = QPushButton(Form)
        self.pbLaunch.setObjectName(u"pbLaunch")

        self.horizontalLayout.addWidget(self.pbLaunch)


        self.verticalLayout_6.addLayout(self.horizontalLayout)


        self.retranslateUi(Form)

        self.sw_observation_type.setCurrentIndex(1)
        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"New observation", None))
        self.label.setText(QCoreApplication.translate("Form", u"Observation id", None))
        self.lb_star.setText(QCoreApplication.translate("Form", u"*", None))
        self.label_8.setText(QCoreApplication.translate("Form", u"Date and time", None))
        self.dteDate.setDisplayFormat(QCoreApplication.translate("Form", u"yyyy-MM-dd hh:mm:ss.zzz", None))
        self.label_9.setText(QCoreApplication.translate("Form", u"Description", None))
        self.cb_time_offset.setText(QCoreApplication.translate("Form", u"Time offset", None))
        self.lbTimeOffset.setText(QCoreApplication.translate("Form", u"Time value", None))
        self.cb_observation_time_interval.setText(QCoreApplication.translate("Form", u"Limit observation to a time interval", None))
        self.label_3.setText(QCoreApplication.translate("Form", u"Independent variables", None))
        ___qtablewidgetitem = self.twIndepVariables.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("Form", u"Variable", None));
        ___qtablewidgetitem1 = self.twIndepVariables.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("Form", u"Type", None));
        ___qtablewidgetitem2 = self.twIndepVariables.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("Form", u"Value", None));
        self.gb_observation_type.setTitle(QCoreApplication.translate("Form", u"Observation type", None))
        self.rb_media_files.setText(QCoreApplication.translate("Form", u"Observation from media file(s)", None))
        self.rb_live.setText(QCoreApplication.translate("Form", u"Live observation", None))
        self.rb_images.setText(QCoreApplication.translate("Form", u"Observation from pictures", None))
        ___qtablewidgetitem3 = self.twVideo1.horizontalHeaderItem(0)
        ___qtablewidgetitem3.setText(QCoreApplication.translate("Form", u"Player", None));
        ___qtablewidgetitem4 = self.twVideo1.horizontalHeaderItem(1)
        ___qtablewidgetitem4.setText(QCoreApplication.translate("Form", u"Offset (seconds)", None));
        ___qtablewidgetitem5 = self.twVideo1.horizontalHeaderItem(2)
        ___qtablewidgetitem5.setText(QCoreApplication.translate("Form", u"Path", None));
        ___qtablewidgetitem6 = self.twVideo1.horizontalHeaderItem(3)
        ___qtablewidgetitem6.setText(QCoreApplication.translate("Form", u"Duration", None));
        ___qtablewidgetitem7 = self.twVideo1.horizontalHeaderItem(4)
        ___qtablewidgetitem7.setText(QCoreApplication.translate("Form", u"FPS", None));
        ___qtablewidgetitem8 = self.twVideo1.horizontalHeaderItem(5)
        ___qtablewidgetitem8.setText(QCoreApplication.translate("Form", u"Video", None));
        ___qtablewidgetitem9 = self.twVideo1.horizontalHeaderItem(6)
        ___qtablewidgetitem9.setText(QCoreApplication.translate("Form", u"Audio", None));
        self.pbAddVideo.setText(QCoreApplication.translate("Form", u"Add media", None))
        self.pbRemoveVideo.setText(QCoreApplication.translate("Form", u"Remove selected media", None))
        self.pb_use_media_file_name_as_obsid.setText(QCoreApplication.translate("Form", u"Use media file name as observation id", None))
        self.cbVisualizeSpectrogram.setText(QCoreApplication.translate("Form", u"Visualize the sound spectrogram for the player #1", None))
        self.cb_visualize_waveform.setText(QCoreApplication.translate("Form", u"Visualize the waveform for the player #1", None))
        self.cb_media_creation_date_as_offset.setText(QCoreApplication.translate("Form", u"Use the media creation date/time as offset", None))
        self.label_5.setText(QCoreApplication.translate("Form", u"Scan sampling every (s)", None))
        self.label_2.setText(QCoreApplication.translate("Form", u"Image display duration (s)", None))
        self.cbCloseCurrentBehaviorsBetweenVideo.setText(QCoreApplication.translate("Form", u"Stop ongoing state events between successive media files", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_player_1), QCoreApplication.translate("Form", u"Media files", None))
        self.label_7.setText(QCoreApplication.translate("Form", u"Data files to plot", None))
        ___qtablewidgetitem10 = self.tw_data_files.horizontalHeaderItem(0)
        ___qtablewidgetitem10.setText(QCoreApplication.translate("Form", u"Path", None));
        ___qtablewidgetitem11 = self.tw_data_files.horizontalHeaderItem(1)
        ___qtablewidgetitem11.setText(QCoreApplication.translate("Form", u"Columns to plot", None));
        ___qtablewidgetitem12 = self.tw_data_files.horizontalHeaderItem(2)
        ___qtablewidgetitem12.setText(QCoreApplication.translate("Form", u"Plot title", None));
        ___qtablewidgetitem13 = self.tw_data_files.horizontalHeaderItem(3)
        ___qtablewidgetitem13.setText(QCoreApplication.translate("Form", u"Variable name", None));
        ___qtablewidgetitem14 = self.tw_data_files.horizontalHeaderItem(4)
        ___qtablewidgetitem14.setText(QCoreApplication.translate("Form", u"Converters", None));
        ___qtablewidgetitem15 = self.tw_data_files.horizontalHeaderItem(5)
        ___qtablewidgetitem15.setText(QCoreApplication.translate("Form", u"Time interval (s)", None));
        ___qtablewidgetitem16 = self.tw_data_files.horizontalHeaderItem(6)
        ___qtablewidgetitem16.setText(QCoreApplication.translate("Form", u"Start position (s)", None));
        ___qtablewidgetitem17 = self.tw_data_files.horizontalHeaderItem(7)
        ___qtablewidgetitem17.setText(QCoreApplication.translate("Form", u"Substract first value", None));
        ___qtablewidgetitem18 = self.tw_data_files.horizontalHeaderItem(8)
        ___qtablewidgetitem18.setText(QCoreApplication.translate("Form", u"Color", None));
        self.pb_add_data_file.setText(QCoreApplication.translate("Form", u"Add data file", None))
        self.pb_remove_data_file.setText(QCoreApplication.translate("Form", u"Remove selected data file", None))
        self.pb_view_data_head.setText(QCoreApplication.translate("Form", u"View data from file", None))
        self.pb_plot_data.setText(QCoreApplication.translate("Form", u"Show plot", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_data_files), QCoreApplication.translate("Form", u"Data files", None))
        self.label_4.setText(QCoreApplication.translate("Form", u"Scan sampling every", None))
        self.label_6.setText(QCoreApplication.translate("Form", u"seconds", None))
        self.cb_start_from_current_time.setText(QCoreApplication.translate("Form", u"Start from current time", None))
        self.rb_day_time.setText(QCoreApplication.translate("Form", u"Day time", None))
        self.rb_epoch_time.setText(QCoreApplication.translate("Form", u"Epoch time (seconds since 1970-01-01)", None))
        self.lb_images_info.setText(QCoreApplication.translate("Form", u"Image info:", None))
        self.pb_add_directory.setText(QCoreApplication.translate("Form", u"Add directory", None))
        self.pb_remove_directory.setText(QCoreApplication.translate("Form", u"Remove directory", None))
        self.pb_use_img_dir_as_obsid.setText(QCoreApplication.translate("Form", u"Use the pictures directory as observation id", None))
        self.groupBox.setTitle(QCoreApplication.translate("Form", u"Time", None))
        self.rb_no_time.setText(QCoreApplication.translate("Form", u"No time", None))
        self.rb_use_exif.setText(QCoreApplication.translate("Form", u"Use the EXIF DateTimeOriginal tag", None))
        self.rb_time_lapse.setText(QCoreApplication.translate("Form", u"Time lapse (s)", None))
        self.pbCancel.setText(QCoreApplication.translate("Form", u"Cancel", None))
        self.pbSave.setText(QCoreApplication.translate("Form", u"Save", None))
        self.pbLaunch.setText(QCoreApplication.translate("Form", u"Start", None))
    # retranslateUi




"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.



Plot events in real time
"""

import matplotlib

matplotlib.use("Qt5Agg")

import numpy as np
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QLabel
from PySide6.QtCore import Signal, QEvent, Qt
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas

from matplotlib.figure import Figure

from . import config as cfg

# matplotlib.pyplot.switch_backend("Qt5Agg")


class Plot_events_RT(QWidget):
    # send keypress event to mainwindow
    sendEvent = Signal(QEvent)

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Events plot")

        self.interval = 60  # default interval of visualization (in seconds)
        self.time_mem = -1

        self.events_mem = {"init": 0}

        self.cursor_color = cfg.REALTIME_PLOT_CURSOR_COLOR  # default cursor color
        self.observation_type = cfg.MEDIA
        self.groupby = "behaviors"  # group results by "behaviors" or "modifiers"

        self.figure = Figure()
        self.ax = self.figure.add_subplot(1, 1, 1)

        self.canvas = FigureCanvas(self.figure)

        layout = QVBoxLayout()
        layout.addWidget(self.canvas)

        hlayout1 = QHBoxLayout()
        hlayout1.addWidget(QLabel("Time interval"))
        hlayout1.addWidget(QPushButton("+", self, clicked=lambda: self.time_interval_changed(1), focusPolicy=Qt.NoFocus))
        hlayout1.addWidget(QPushButton("-", self, clicked=lambda: self.time_interval_changed(-1), focusPolicy=Qt.NoFocus))
        self.pb_mode = QPushButton("Include modifiers", self, clicked=self.change_mode, focusPolicy=Qt.NoFocus)
        hlayout1.addWidget(self.pb_mode)
        layout.addLayout(hlayout1)

        self.setLayout(layout)

        self.installEventFilter(self)

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.sendEvent.emit(event)
            return True
        else:
            return False

    def change_mode(self) -> None:
        """
        Change plot mode
        "behaviors" -> plot behaviors without modifiers
        "modifiers" -> plot behaviors and modifiers
        """

        if self.groupby == "behaviors":
            self.groupby = "modifiers"
            self.pb_mode.setText("Show behaviors w/o modifiers")
        else:
            self.groupby = "behaviors"
            self.pb_mode.setText("Include modifiers")

    def time_interval_changed(self, action: int) -> None:
        """
        change the time interval for events plot

        Args:
            action (int): -1 decrease time interval, +1 increase time interval

        Returns:
            None
        """

        if action == -1 and self.interval <= 5:
            return
        self.interval += 5 * action
        self.plot_events(current_time=self.time_mem, force_plot=True)

    def aggregate_events(self, events: list, start: float, end: float) -> dict:
        """
        aggregate state events
        take consideration of subject and modifiers

        Args:
            events (list): list of events
            start (float): initial value
            end (float): final value

        Returns:
            dict

        """

        def group(subject: str, code: str, modifier: str) -> tuple:
            if self.groupby == "behaviors":
                return (subject, code)
            else:  # with modifiers
                return (subject, code, modifier)

        mem_behav = {}
        intervals_behav = {}

        for event in events:
            intervals_behav[group(event[1], event[2], event[3])] = [(0, 0)]

        for event in events:
            time_, subject, code, modifier = event[:4]
            key = group(subject, code, modifier)

            # check if code is state
            if code in self.state_events_list:
                if key in mem_behav and mem_behav[key] is not None:
                    # stop interval

                    # check if event is in interval start-end
                    if any(
                        (
                            start <= mem_behav[key] <= end,
                            start <= time_ <= end,
                            mem_behav[key] <= start and time_ > end,
                        )
                    ):
                        intervals_behav[key].append((float(mem_behav[key]), float(time_)))
                    mem_behav[key] = None
                else:
                    # start interval
                    mem_behav[key] = time_

            else:  # point event
                if start <= time_ <= end:
                    intervals_behav[key].append((float(time_), float(time_) + self.point_event_plot_duration * 50))  # point event -> 1 s

        # check if intervals are closed
        for k in mem_behav:
            if mem_behav[k] is not None:  # interval open
                if self.observation_type == cfg.LIVE:
                    intervals_behav[k].append((float(mem_behav[k]), float((end + start) / 2)))  # close interval with current time

                elif self.observation_type == cfg.MEDIA:
                    intervals_behav[k].append((float(mem_behav[k]), float(end)))  # close interval with end value

        return intervals_behav

    def plot_events(self, current_time: float, force_plot: bool = False):
        """
        plot events centered on the current time

        Args:
            current_time (float): time for displaying events
            force_plot (bool): force plot even if media paused
        """

        self.events = self.aggregate_events(self.events_list, current_time - self.interval / 2, current_time + self.interval / 2)

        if not force_plot and current_time == self.time_mem:
            return

        self.time_mem = current_time

        if self.events != self.events_mem:
            left, duration = {}, {}
            for k in self.events:
                left[k] = []
                duration[k] = []
                for interv in self.events[k]:
                    left[k].append(interv[0])
                    duration[k].append(interv[1] - interv[0])

            self.behaviors, self.durations, self.lefts, self.colors = [], [], [], []
            for k in self.events:
                if self.groupby == "behaviors":
                    subject_name, bevavior_code = k
                    if subject_name == "":
                        subject_name = "No focal"
                    behav_col = self.behav_color[bevavior_code]
                    self.behaviors.extend([f"{subject_name} - {bevavior_code}"] * len(self.events[k]))
                    self.colors.extend([behav_col] * len(self.events[k]))
                else:  # with modifiers
                    subject_name, bevavior_code, modifier = k
                    behav_col = self.behav_color[bevavior_code]
                    self.behaviors.extend([f"{subject_name} - {bevavior_code} ({modifier})"] * len(self.events[k]))
                    self.colors.extend([behav_col] * len(self.events[k]))

                self.lefts.extend(left[k])
                self.durations.extend(duration[k])

            self.events_mem = self.events

        self.ax.clear()
        self.ax.set_xlim(current_time - self.interval / 2, current_time + self.interval / 2)

        self.ax.axvline(x=current_time, color=self.cursor_color, linestyle="-")

        self.ax.barh(y=np.array(self.behaviors), width=self.durations, left=self.lefts, color=self.colors, height=0.5)

        self.canvas.draw()
        self.figure.canvas.flush_events()



"""
module for testing time budget analysis

pytest -s -vv test_irr.py
"""

import pytest
import sys
import json
import os
import decimal


sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from boris import irr
from boris import db_functions
from boris import config 


class Test_irr(object):

    def test_cohen_kappa_same_observation(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #1", "observation #1"]
        selected_subjects = ["subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s", "p"])
        assert r == True

        cursor = db.cursor()
        K, msg = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)


        assert K == 1


    def test_cohen_kappa_very_different_obs(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #1", "observation #2"]
        selected_subjects = ["subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        K, msg = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)


        assert K == -0.036


    def test_cohen_kappa_very_similar_obs_interval_state_1s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        K, _ = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        assert K == 0.988


    def test_cohen_kappa_very_similar_obs_interval_state_3s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        K, _ = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("3"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        assert K == 1


    def test_cohen_kappa_very_similar_obs_interval_point_1s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["p"])

        cursor = db.cursor()
        K, _ = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        assert K == 0.866


    def test_cohen_kappa_very_similar_obs_interval_point_3s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["p"])

        cursor = db.cursor()
        K, _ = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("3.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        assert K == 0.798


    def test_obs_without_events(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation without events"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        K, _ = irr.cohen_kappa(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        assert K == -100


    def test_needleman_wunsch_identity_same_observation(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #1", "observation #1"]
        selected_subjects = ["subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s", "p"])
        assert r == True

        cursor = db.cursor()
        identity, msg = irr.needleman_wunsch_identity(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)



        assert identity == 100.0


    def test_needleman_wunsch_identity_very_different_obs(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #1", "observation #2"]
        selected_subjects = ["subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        identity, msg = irr.needleman_wunsch_identity(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        # print(identity)
        assert identity == 85.84070796460178


    def test_needleman_wunsch_identity_very_similar_obs_state_1s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        identity, msg = irr.needleman_wunsch_identity(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        # print(identity)
        assert identity == 99.36708860759494


    def test_needleman_wunsch_identity_very_similar_obs_state_3s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["s"])

        cursor = db.cursor()
        identity, msg = irr.needleman_wunsch_identity(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("3.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        # print(identity)
        assert identity == 100.0


    def test_needleman_wunsch_identity_very_similar_obs_point_1s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["p"])

        cursor = db.cursor()
        identity, msg = irr.needleman_wunsch_identity(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("1.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        # print(identity)
        assert identity == 99.29577464788733

    def test_needleman_wunsch_identity_very_similar_obs_point_3s(self):

        pj = json.loads(open("files/test.boris").read())

        ethogram = pj[config.ETHOGRAM]
        selected_observations = ["observation #2", "observation #2 (copy)"]
        selected_subjects = ["No focal subject", "subject1", "subject2"]

        r, s, db = cursor = db_functions.load_aggregated_events_in_db(pj,
                                              selected_subjects,
                                              selected_observations,
                                              ["p"])

        cursor = db.cursor()
        identity, msg = irr.needleman_wunsch_identity(cursor,
                               obsid1 = selected_observations[0],
                               obsid2 = selected_observations[1],
                               interval = decimal.Decimal("3.0"),
                               selected_subjects = selected_subjects,
                               include_modifiers = False)

        print(identity)
        assert identity == 97.91666666666666



"""

widget to edit duration > 24 h or < 0

https://stackoverflow.com/questions/44380202/creating-a-custom-widget-in-PySide6
"""

from decimal import Decimal as dec

from PySide6.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QPushButton,
    QSpinBox,
    QLabel,
    QSpacerItem,
    QSizePolicy,
    QDoubleSpinBox,
    QStackedWidget,
    QRadioButton,
)

from PySide6.QtCore import Signal

from . import config as cfg


class Widget_hhmmss(QWidget):
    time_changed_signal = Signal(float)

    def __init__(self, parent=None):
        QWidget.__init__(self, parent=parent)

        lay = QHBoxLayout(self)
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)

        self.sign = QPushButton("+")
        self.sign.clicked.connect(self.change_sign)
        self.sign.setStyleSheet("font-size:12px;")
        lay.addWidget(self.sign)

        self.hours = QSpinBox()
        self.hours.setValue(0)
        self.hours.setMinimum(0)
        self.hours.setMaximum(2**31 - 1)
        self.hours.valueChanged.connect(self.update_time_value)
        lay.addWidget(self.hours)
        lay.addWidget(QLabel(":"))

        self.minutes = QSpinBox()
        self.minutes.setMinimum(-1)
        self.minutes.setMaximum(60)
        self.minutes.setPrefix("0")
        self.minutes.valueChanged.connect(lambda value, x=self.minutes: self.value_changed(value, x, 0, 59))
        lay.addWidget(self.minutes)
        lay.addWidget(QLabel(":"))

        self.seconds = QSpinBox()
        self.seconds.setMinimum(-1)
        self.seconds.setMaximum(60)
        self.seconds.setPrefix("0")
        self.seconds.valueChanged.connect(lambda value, x=self.seconds: self.value_changed(value, x, 0, 59))
        lay.addWidget(self.seconds)
        lay.addWidget(QLabel(":"))

        self.milliseconds = QSpinBox()
        self.milliseconds.setMinimum(-1)
        self.milliseconds.setMaximum(1000)
        self.milliseconds.setPrefix("00")
        self.milliseconds.valueChanged.connect(lambda value, x=self.milliseconds: self.value_changed(value, x, 0, 999))
        lay.addWidget(self.milliseconds)

        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        lay.addItem(spacerItem)
        self.setLayout(lay)

    def change_sign(self):
        """
        sign has changed
        """
        self.sign.setText("+" if self.sign.text() == "-" else "-")
        self.update_time_value()

    def update_time_value(self):
        new_time = self.hours.value() * 3600 + self.minutes.value() * 60 + self.seconds.value() + self.milliseconds.value() / 1000
        if self.sign.text() == "-":
            new_time = -new_time

        self.time_changed_signal.emit(new_time)

    def value_changed(self, new_value, widget, val_min, val_max):
        """
        time value changed
        """

        if new_value > val_max:
            widget.setValue(val_min)
            widget.setPrefix("0" if val_max < 100 else "00")
            self.update_time_value()
            return

        if new_value < val_min:
            widget.setValue(val_max)
            widget.setPrefix("")
            self.update_time_value()
            return

        if new_value < 10:
            widget.setPrefix("0" if val_max < 100 else "00")
        elif new_value < 100:
            widget.setPrefix("" if val_max < 100 else "0")
        else:
            widget.setPrefix("")

        self.update_time_value()


class Widget_seconds(QWidget):
    time_changed_signal = Signal(float)

    def __init__(self, parent=None):
        QWidget.__init__(self, parent=parent)
        lay = QHBoxLayout(self)
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)

        self.seconds2 = QDoubleSpinBox()
        self.seconds2.setValue(0)
        self.seconds2.setMinimum(-100_000_000)
        self.seconds2.setMaximum(100_000_000)
        self.seconds2.setDecimals(3)
        self.seconds2.valueChanged.connect(self.value_changed)
        lay.addWidget(self.seconds2)

        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        lay.addItem(spacerItem)

        self.setLayout(lay)

    def value_changed(self, v):
        self.time_changed_signal.emit(v)


class Duration_widget(QWidget):
    def __init__(self, time_value=0, parent=None):
        super().__init__()

        self.time_value = dec(time_value).quantize(dec(".001"))

        lay = QHBoxLayout(self)
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)

        self.Stack = QStackedWidget()

        self.w1 = Widget_hhmmss()
        self.w1.time_changed_signal.connect(self.time_changed)
        self.Stack.addWidget(self.w1)

        self.w2 = Widget_seconds()
        self.w2.time_changed_signal.connect(self.time_changed)
        self.Stack.addWidget(self.w2)

        lay.addWidget(self.Stack)

        self.format_hhmmss = QRadioButton("HH:MM:SS:MS")
        self.format_hhmmss.setChecked(True)
        self.format_hhmmss.clicked.connect(self.set_format_hhmmss)
        lay.addWidget(self.format_hhmmss)
        self.format_s = QRadioButton("seconds")
        self.format_s.clicked.connect(self.set_format_s)
        lay.addWidget(self.format_s)

        lay.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        self.setLayout(lay)

        self.set_time(self.time_value)

    def time_changed(self, x):
        """
        widget time has changed
        """
        self.time_value = x

    def set_time(self, new_time):
        if new_time.is_nan():
            return

        self.w1.sign.setText("-" if new_time < 0 else "+")

        h = int(abs(new_time) // 3600)
        m = int((abs(new_time) - h * 3600) // 60)
        s = int((abs(new_time) - h * 3600 - m * 60))
        ms = round((abs(new_time) - h * 3600 - m * 60 - s) * 1000)

        self.w1.hours.setValue(h)
        self.w1.minutes.setValue(m)
        self.w1.seconds.setValue(s)
        self.w1.milliseconds.setValue(ms)

        self.w2.seconds2.setValue(new_time)

        self.time_value = new_time

    def set_format_s(self):
        """
        switch to seconds widget
        """

        self.format_s.setChecked(True)
        self.w2.seconds2.setValue(self.time_value)
        self.Stack.setCurrentIndex(1)

    def set_format_hhmmss(self):
        """
        switch to HHMMSS widget
        """

        self.format_hhmmss.setChecked(True)
        self.set_time(self.time_value)
        self.Stack.setCurrentIndex(0)

    def set_format(self, time_format):
        """
        switch time format in base of time_format value
        """
        if time_format in [cfg.HHMMSS, cfg.HHMMSSZZZ]:
            self.set_format_hhmmss()
        if time_format in [cfg.S]:
            self.set_format_s()

    def get_time(self) -> dec:
        """
        return time displayed by widget in seconds
        """
        return dec(self.time_value).quantize(dec(".001"))
